#
# sbackup - legacy support for restore operation
#
#   Copyright (c)2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2005-2008: Aigars Mahinovs <aigarius@debian.org>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import shutil
import datetime
import tempfile
import filecmp

import sys
import os.path
import re
import threading
import time
import traceback

import sbackup.util as Util

# Attempt to load GTK bindings
try:
    import pygtk
    pygtk.require("2.0")
    import gtk
    import gtk.glade
except ImportError:
    print "Failed to load Python GTK/Gnome bindings. Please check your Gnome installation."
    sys.exit(1)
try:
    import gnomevfs
except ImportError:
    import gnome.vfs as gnomevfs
    
from sbackup.ui import misc
sys.excepthook = misc.except_hook_threaded


_LEGACY_RESTOREGUI_GLADE = Util.get_resource_file('sbackup-legacy-restore.glade')


class SRestore:
    def __init__(self):
        pass

    def restore( self, backup, spath, dpath = None ):
        """
        Restore one file or directory from the backup tdir with name
        spath to dpath (or to its old location).
        All existing files must be moved to a "*.before_restore_$time" files.
        """
        
        if not dpath:
            dpath = spath

        # Gather spath and dpath information
        if spath[0] == "/": spath = spath[1:]
        (sparent, sname) = os.path.split( spath )
        if not sname:
            spath = sparent
            (sparent, sname) = os.path.split( sparent )
        dpath = os.path.normpath( dpath )
        (dparent, dname) = os.path.split( dpath )
        if not dname:
            dpath = dparent
            (dparent, dname) = os.path.split( dpath )
        
        now = datetime.datetime.now().isoformat("_").replace( ":", "." )
        ver = str(gnomevfs.read_entire_file( backup+"/ver" ))
        
        try: 
            if ver[:3] == "1.4":
                self.childlist = [x[1:] for x in gnomevfs.read_entire_file( backup+"/flist" ).split( "\000" ) if x == "/"+spath or x[1:len(spath)+2]==spath+"/"]
            else:
                print "Only snapshot version 1.4 is supported"
                return False
        except:
            print "E: Error opening backup snapshot metadata"
            return False

        if len(self.childlist) == 0:
            print "E: File not found in the backup snapshot"
            return False

        if os.path.exists(dpath):
            if os.path.isdir(dpath):
                tdir = tempfile.mkdtemp( dir=dpath )
                self.extract( backup, spath, tdir )
                for _file in self.childlist:
                    if len(self.childlist)==1:
                        bname = sname
                        src = os.path.join( tdir, spath )
                    else:
                        bname = _file[len(spath)+1:]
                        src = os.path.join( tdir, spath, bname )
                    dst = os.path.join( dpath, bname )
                    if os.path.isdir(src):
                        if not os.path.exists(dst):
                            os.makedirs(dst)
                        srcstat = os.stat( src )
                        os.chown( dst, srcstat.st_uid, srcstat.st_gid )
                        os.chmod( dst, srcstat.st_mode )
                    elif os.path.isfile(src) or os.path.islink(src):
                        if os.path.exists(dst) and not filecmp.cmp(src, dst):
                            shutil.move( dst, dst+".before_restore_"+now )
                        if not os.path.exists(dst):
                            shutil.move( src, dst )
                    else:
                        print "W: Path '%s' is neither directory nor file! Skipped." % src
                shutil.rmtree( tdir )
            else:
                tdir = tempfile.mkdtemp( dir=dparent )
                self.extract( backup, spath, tdir )
                shutil.move( dpath, dpath+".before_restore_"+now )
                shutil.move( os.path.join(tdir,spath), dpath )
                shutil.rmtree( tdir )
                
        else:
            tdir = tempfile.mkdtemp( dir=dparent )
            self.extract( backup, spath, tdir )
            shutil.move( os.path.join(tdir,spath), dpath )
            shutil.rmtree( tdir )

        return True

    def islocal( self, uri ):
        local = True
        try:
            if not gnomevfs.URI( uri ).is_local:
                local = False
        except:
            pass
        return local

    def extract( self, backup, spath, tdir ):
        tarline = "tar -xzp --occurrence=1 --ignore-failed-read -C '"+tdir+"' "
        if self.islocal( backup ):
            tarline += " -f '"+backup+"/files.tgz' '"+spath+"' >/dev/null 2>&1"
            os.system( tarline )
        else:
            tarline += "'"+spath+"' 2>/dev/null"
            tsrc = gnomevfs.open( backup+"/files.tgz", 1)
            tdst = os.popen( tarline, "w" )
            try: shutil.copyfileobj( tsrc, tdst, 100*1024 )
            except gnomevfs.EOFError: pass
            tdst.close()
            tsrc.close()


class SRestoreGTK:
    """
    Main GUI application class.
    """
    
    default_target = "/var/backup"
    target = default_target
    versions = []
    
    
    def __init__(self):
        """
        Initializes the application.
        """

        # Setup glade and signals
        gtk.glade.textdomain("sbackup")
        self.signals = {"gtk_main_quit": gtk.main_quit,
                "on_customsrc_toggled": self.enable_custom,
                "on_treeview1_row_expanded": self.on_expand_row,
                "on_backup_changed": self.on_backup_changed,
                "on_apply_clicked": self.on_custom_apply,
                "on_move_cursor":self.on_selection_change,
                "on_restore":self.restore,
                "on_restore_as":self.restore_as,
                "on_customFolderButton_clicked": self.on_customFolderButton_clicked
                }

        self.widgets = gtk.glade.XML(_LEGACY_RESTOREGUI_GLADE)
        
        self.widgets.signal_autoconnect(self.signals)
        
        # Get handle to window
        self.window = self.widgets.get_widget("restore")

        self.widgets.get_widget("labelDefaultSource").set_text(self.default_target)    

        # Load the backup tree from the default location
        self.init_tree()
        self.sel = self.flist_widget.get_selection()
        self.sel.set_mode( gtk.SELECTION_SINGLE )
        self.load_tree(self.default_target)
        
        gtk.gdk.threads_init()
        # Start the main loop
        gtk.main()
    
    def init_tree(self):
        """
        Initalizes the tree structure
        """
        
        self.flist_widget = self.widgets.get_widget("treeview1")
        self.treestore = gtk.TreeStore( str )
        
        self.flist_widget.set_model( self.treestore )
        
        acolumn = gtk.TreeViewColumn( "Path", gtk.CellRendererText(), text=0 )
        self.flist_widget.append_column( acolumn )

        blist_widget = self.widgets.get_widget( "combobox1" )
        self.blist = gtk.ListStore(str)
        blist_widget.set_model( self.blist )
        cell = gtk.CellRendererText()
        blist_widget.pack_start( cell, True )
        blist_widget.add_attribute( cell, "text", 0)

        
    def load_tree(self, target):
        """
        Loads the tree information from the target backup directory
        """
        self.treestore.clear()
        
        # Checking if the target directory is local or remote
        local = True
        try:
            if gnomevfs.URI( target ).is_local:
                target = gnomevfs.get_local_path_from_uri( target )
            else:
                local = False
        except:
            pass

        self.local = local
        self.target = target
        
        # Checking if it is a readable directory
        if local:
            if not (os.path.exists( target ) and os.path.isdir( target ) and os.access( target, os.R_OK | os.X_OK ) ):    
                self.treestore.append( None, ["Error: backups directory does not exist!"])
                self.target = False
        else:
            if not (gnomevfs.exists( target ) and gnomevfs.get_file_info(target).type == 2):
                self.treestore.append( None, ["Error: backups directory does not exist!"])
                self.target = False
        
        # Get list of backup directories
        r = re.compile(r"^(\d{4})-(\d{2})-(\d{2})_(\d{2})[\:\.](\d{2})[\:\.](\d{2})\.\d+\..*?\.(.+)$")
        
        listing = []
    
        if local and self.target:
            listing = os.listdir( target )
            listing = filter( r.search, listing )
        elif self.target:
            try:     
                d = gnomevfs.open_directory( target )
                listing = []
                for f in d:
                    if f.type == 2 and f.name != "." and f.name != ".." and r.search( f.name ):
                        listing.append( f.name )
            except: pass

        # Check if these directories are complete and remove from the list those that are not
        for adir in listing[:]:
            if not gnomevfs.exists(self.target+"/"+adir+"/ver") :
                listing.remove( adir )
                continue
            else :
                if str(gnomevfs.read_entire_file( self.target+"/"+adir+"/ver"))[:3] != "1.4":
                    listing.remove( adir )
                    continue

        listing.sort()
        listing.reverse()

        self.vtree = {}

        if listing == []:
            self.treestore.append( None, ["Error: no backups found in the target directory"])
            self.target = False
        else:
            for base in listing:
                if str(gnomevfs.read_entire_file(target+"/"+base+"/ver"))[:3] == "1.4":
                    self.vtree[base] = str(gnomevfs.read_entire_file(target+"/"+base+"/flist")).split("\000")
                else:
                    self.vtree[base] = str(gnomevfs.read_entire_file(target+"/"+base+"/flist")).split("\n")
        
        self.blist.clear()
        
        for base in listing:
            self.blist.append( [base] )
        
        self.good = False
        self.on_selection_change()
        if self.target:
            self.treestore.append( None, ["Select any of the available backups to see list of files that can be restored."])        
        
    def on_customFolderButton_clicked(self, *args):
        dialog = gtk.FileChooserDialog("Choose a source folder", None, gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        dialog.set_local_only(False)
        if dialog.run() == gtk.RESPONSE_OK:
            self.widgets.get_widget("entry1").set_text(dialog.get_uri())
        dialog.destroy()

    def enable_custom(self, *args):
        """
        Enables/Disables input box for the custom backup dir
        Reloads default dir on disabling
        """
        if self.widgets.get_widget("radiobutton2").get_active():
            self.widgets.get_widget("entry1").set_sensitive(True)
            self.widgets.get_widget("customFolderButton").set_sensitive(True)
            self.widgets.get_widget("button7").set_sensitive(True)
            self.widgets.get_widget("labelDefaultSource").set_sensitive(False)
        else:
            self.widgets.get_widget("entry1").set_sensitive(False)
            self.widgets.get_widget("customFolderButton").set_sensitive(False)
            self.widgets.get_widget("button7").set_sensitive(False)
            self.widgets.get_widget("labelDefaultSource").set_sensitive(True)
            self.load_tree(self.default_target)

    def on_backup_changed( self, combox ):
        """
        Reset the file tree view.
        """
        self.treestore.clear()
        self.treestore.append( None, ["dummy"])
        self.show_dir( "", None )
        
    def on_expand_row( self, tv, aiter, path, user_data=None):
        """
        When a row in the file tree view is expanded, we populate
        it with children (unless they are there already).
        """
        if self.treestore.iter_nth_child( aiter, 1 ):
            return
        self.show_dir( self.path_to_dir(path), aiter )

    def path_to_dir( self, path ):
        """
        Recievs path in the treestore (as tuple) and returns a directory
        path as string.
        """
        g = list(path)
        p = ""
        while g != []:
            i = self.treestore.get_iter( tuple(g) )
            p = "/" + self.treestore.get_value( i, 0 ) + p
            g = g[:-1]
        return p
    
    def show_dir(self, path, rootiter):
        """
        Worker function - adds all files/directories from the filez list
        to the treestore at the rootiter.
        """
        dummy = self.treestore.iter_children(rootiter)
        
        self.good = True

        base = self.get_active_text(self.widgets.get_widget("combobox1"))
        list2 = []
        list3 = []
        
        escapedFullPath = re.escape(path)+"/([^/]+/?)"
        for item in self.vtree[base]:
            m = re.match( escapedFullPath, item )
            if m and not list2.count(m.group(1)) and not list3.count(m.group(1)[:-1]):
                if m.group(1)[-1] == "/":
                    if list2.count(m.group(1)[:-1]):
                        list2.remove(m.group(1)[:-1])
                    list3.append(m.group(1)[:-1])
                else:
                    list2.append( m.group(1) )
        for d in list3:
            aiter = self.treestore.append( rootiter, [d] )
            self.treestore.append( aiter, ["Loading ..."] )
        for f in list2:
            self.treestore.append( rootiter, [f] )
        
        self.treestore.remove( dummy )
        
    def get_active_text(self, combobox):
        model = combobox.get_model()
        active = combobox.get_active()
        if active < 0:
            return None
        return model[active][0]
    
    def on_custom_apply(self, *args):
        """
        Reload all backup info from a custom location
        """
        
        self.load_tree(self.widgets.get_widget("entry1").get_text())

    def on_selection_change(self, *args):
        """
        Enable/disable restore buttons as selection changes
        """
        (model, aiter) = self.sel.get_selected()
        if aiter and self.good:
            self.widgets.get_widget("button2").set_sensitive(True)
            self.widgets.get_widget("button3").set_sensitive(True)
        else:
            self.widgets.get_widget("button2").set_sensitive(False)
            self.widgets.get_widget("button3").set_sensitive(False)
    
    def _restore_init( self, *args):
        """
        Internal function to prepare for restorin a file
        """
        (store, aiter) = self.widgets.get_widget("treeview1").get_selection().get_selected()
        self.src = self.path_to_dir( store.get_path( aiter ) )
        return aiter
    
    def _do_restore( self, src, dst):
        """ Internal function to ask for confirmation and call the real restore library func"""
        dialog = gtk.MessageDialog(parent=None, flags=0,
                        type=gtk.MESSAGE_QUESTION, buttons=gtk.BUTTONS_YES_NO,
                        message_format="Do you really want to restore backuped copy of '%s' to '%s' ?" % (src, dst))
        response = dialog.run()
        dialog.destroy()
        if response == gtk.RESPONSE_YES:
            while gtk.events_pending():
                gtk.main_iteration(False)
            dialog = self.widgets.get_widget("restore_progress_dialog")
            dialog.show()

            progressBar = self.widgets.get_widget("progressbar")
            progressThread = ProgressThread(dialog, progressBar)
            progressThread.setDaemon(True)
            progressThread.start()
            tdir = self.target+"/"+self.get_active_text(self.widgets.get_widget("combobox1"))
            self.restoreThread = RestoreThread(tdir, src, dst, progressThread)
            self.restoreThread.setDaemon(True)
            self.restoreThread.start()

    def restore( self, *args):
        """ Restore selected path to its original location"""
        self._restore_init()
        self._do_restore( self.src, self.src )

    def restore_as( self, *args):
        """ Restore selected path to a specific location"""
        aiter = self._restore_init()
        if self.treestore.iter_children( aiter ):
            # is a directory
            dialog = gtk.FileChooserDialog(title="Select restore location",
                                action=gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                         gtk.STOCK_OPEN,gtk.RESPONSE_OK))
            dialog.set_filename( self.src )
            result = dialog.run()
            filename = dialog.get_filename()
            dialog.destroy()

            if result == gtk.RESPONSE_OK:
                self._do_restore( self.src, filename )
        else:
            dialog = gtk.FileChooserDialog(title="Select restore location",
                                action=gtk.FILE_CHOOSER_ACTION_SAVE,
                                buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                         gtk.STOCK_OPEN,gtk.RESPONSE_OK))
            dialog.set_filename( self.src )
            dialog.set_current_name( self.src )
            result = dialog.run()
            filename = dialog.get_filename()
            dialog.destroy()

            if result == gtk.RESPONSE_OK:
                self._do_restore( self.src, filename )


class RestoreThread(threading.Thread):

    def __init__(self, tdir, src, dst, progressThread):
        threading.Thread.__init__(self)
        self.tdir = tdir
        self.src = src
        self.dst = dst
        self.alive = True
        self.progressThread = progressThread

    def run(self):
        try:
            r = SRestore()
            r.restore( self.tdir, self.src, self.dst )
            del r
        except:
            _exc = traceback.format_exc()
            misc.show_errdialog_threaded(message_str = "An uncaught error "\
                        "occurred. Close this message window and restart the "\
                        "application.\n\nPlease report this error on "\
                        "https://bugs.launchpad.net/sbackup.",
                        parent = None,
                        headline_str = "Sorry, this should not have happened",
                        secmsg_str = _exc)
        self.alive = False
        self.progressThread.cancel = True


class ProgressThread(threading.Thread):

    def __init__(self, dialog, progressBar):
        threading.Thread.__init__(self)
        self.dialog = dialog
        self.progressBar = progressBar
        self.cancel = False

    def run(self):
        while not self.cancel:
            self.progressBar.pulse()
            time.sleep(0.5)

        self.dialog.destroy()



#   Simple Backup - GTK+ GUI resources
#
#   Copyright (c)2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


CONFIGGUI_TOPWINDOW = 'sbackupConfApp'
CONFIGGUI_GLADEFILE = 'sbackup-config.glade'

RESTOREGUI_TOPWINDOW = 'restorewindow'
RESTOREGUI_GLADEFILE = 'sbackup-restore.glade'


def get_configgui_widgets():
    widget_list = [
        'sbackupConfApp',
        'menubar',
        'toolbar',
        'askSaveDialog',
        'fusehelplabel',
        'regexdialog',
        'ftypedialog',
        'dialog-vbox4',
        'vbox18',
        'ftype_st',
        'ftype_box',
        'hbox25',
        'ftype_custom',
        'ftype_custom_ex',
        'cancelbutton',
        'okbutton',
        'dialog-vbox5',
        'vbox19',
        'regex_box',
        'cancelbutton2',
        'okbutton2',
        'statusBar',
        'vbox17',
        'save',
        'save_as',
        'saveButton',
        'exit',
        'imagemenuitem6',
        'imagemenuitem7',
        'imagemenuitem8',
        'imagemenuitem9',
        'about',
        'vbox1',
        'notebook',
#
# general/main page
        'label_general_page',
        'vbox_general_page',
        'cformat',
        'splitsizeCB',
        'splitsizeSB',
        'splitsizevbox',
        'label_splitsize_custom',
#
        'vbox3',
        'scrolledwindow1',
        'includetv',
        'inc_addfile',
        'hbox4',
        'inc_adddir',
        'hbox5',
        'inc_del',
        'notebook2',
        'vbox4',
        'scrolledwindow2',
        'ex_pathstv',
        'ex_addfile1',
        'hbox6',
        'ex_adddir',
        'hbox7',
        'ex_delpath',
        'vbox5',
        'scrolledwindow3',
        'ex_ftypetv',
        'ex_addftype',
        'ex_delftype',
        'vbox6',
        'scrolledwindow4',
        'ex_regextv',
        'ex_addregex',
        'ex_delregex',
        'vbox7',
        'hbox8',
        'ex_max',
        'ex_maxsize',
        'vbox_destination_page',
        'dest2',
        'hbox9',
        'dest_local',
        'btn_browse_local',
        'dest3',
        'hbox10',
        'dest_remote',
        'dest_remote_light',
        'hbox11',
        'dest_unusable',
#
#            schedule page
        'label_schedule_page',
        'vbox_schedule_page',
        'table_schedule',
        'rdbtn_no_schedule',
        'rdbtn_simple_schedule',
        'rdbtn_custom_schedule',
        'label_simple_schedule_freq',
        'label_custom_cronline',
        'txtfld_custom_cronline',
        'cmbbx_simple_schedule_freq',
        'lnkbtn_help_schedule',
        'hbox_schedule_footer',
        'hbox_schedule_infotext',
        'img_schedule_infotext',
        'label_schedule_infotext',
#
        'time_maxinc',
        'purgevbox',
        'purgecheckbox',
        'hbox16',
        'purgeradiobutton',
        'purgedays',
        'purgelabel',
        'hbox17',
        'logpurgeradiobutton',
        'purgelabel2',
        'hbox18',
        'reportvbox',
        'hbox19',
        'table2',
        'loglevelcombobox',
        'logfilechooser',
        'hbox20',
        'vbox11',
        'vbox12',
        'table3',
        'smtpport',
        'smtpserver',
        'smtpto',
        'smtpfrom',
        'hbox21',
        'testMailButton',
        'vbox13',
        'vbox14',
        'smtplogincheckbox',
        'smtplogininfo',
        'smtppassword',
        'smtplogin',
        'vbox15',
        'TLScheckbutton',
        'TLSinfos',
        'TLSradiobutton',
        'SSLradiobutton',
        'crtfilechooser',
        'keyfilechooser',
        'label_certificate',
        'label_key',
#
        'pluginscombobox',
        'ProfileManagerDialog',
        'profilesListTreeView',
        'addProfileButton',
        'removeProfileButton',
        'editProfileButton',
        'closeProfileManagerButton',
        'askNewPrfNameDialog',
        'enableNewPrfCB',
        'newPrfNameEntry',
        'followlinks',
# dialog set default settings
        'dialog_default_settings',
        'label_dialog_default_settings_content',
        'btn_set_default_settings',
        'btn_cancel_default_settings',
#
# dialog set remote destination
        'dialog_connect_remote',
        'btn_set_remote',
        'btn_cancel_remote',
        'btn_connect_remote',
        'cmb_set_remote_service',
        'entry_set_remote_server',
        'entry_set_remote_port',
        'entry_set_remote_dir',
        'entry_set_remote_user',
        'entry_set_remote_pass',
        'label_set_remote_user',
        'label_set_remote_pass',
        'checkbtn_show_password',
# dialog make backup
        'dialog_make_backup',
        'checkbtn_make_backup_full',
        'btn_make_backup',
        'btn_cancel_make_backup',
        'dialog_browse_localdest'
        ]
    return widget_list


def get_configgui_handlers():
    handlers = [
        'on_ftype_toggled',
        'on_ftype_st_box_changed',
        'on_ftype_custom_ex_changed',
        'on_save_activate',
        'on_save_as_activate',
        'on_exit_activate',
        'on_prfManager_activate',
        'on_menu_about_activate',
        'on_menu_help_activate',
        'on_reload_clicked',
        'on_save_clicked',
        'on_backup_clicked',
        'on_cformat_changed',
        'on_splitsizeCB_changed',
        'on_splitsizeSB_value_changed',
        'on_inc_addfile_clicked',
        'on_inc_adddir_clicked',
        'on_inc_del_clicked',
        'on_ex_addfile_clicked',
        'on_ex_adddir_clicked',
        'on_ex_delpath_clicked',
        'on_ex_addftype_clicked',
        'on_ex_delftype_clicked',
        'on_ex_addregex_clicked',
        'on_ex_delregex_clicked',
        'on_ex_max_toggled',
        'on_ex_maxsize_changed',
        'on_dest_toggled',
        'on_btn_browse_local_clicked',
#     connect to remote host
        'on_btn_set_remote_clicked',
        'on_cmb_set_remote_service_changed',
        'on_checkbtn_show_password_toggled',
#     scheduling
        'on_cmbbx_simple_schedule_freq_changed',
        'on_rdbtn_schedule_toggled',
        'on_txtfld_custom_cronline_changed',
#
        'on_time_maxinc_changed',
        'on_purgecheckbox_toggled',
        'on_purgeradiobutton_toggled',
        'on_purgedays_changed',
        'on_logfilechooser_selection_changed',
        'on_loglevelcombobox_changed',
        'on_smtpfrom_changed',
        'on_smtpto_changed',
        'on_smtpserver_changed',
        'on_smtpport_changed',
        'on_testMailButton_clicked',
        'on_smtplogincheckbox_toggled',
        'on_smtplogin_changed',
        'on_smtppassword_changed',
        'on_TLScheckbutton_toggled',
        'on_TLSradiobutton_toggled',
        'on_crtfilechooser_selection_changed',
        'on_keyfilechooser_selection_changed',
        'on_addProfileButton_clicked',
        'on_removeProfileButton_clicked',
        'on_editProfileButton_clicked',
        'on_closeProfileManagerButton_clicked',
        'on_includetv_key_press_event',
        'on_ex_pathstv_key_press_event',
        'on_ex_ftypetv_key_press_event',
        'on_ex_regextv_key_press_event',
        'on_followlinks_toggled',
#
        'on_menu_set_default_settings_activate',

        'on_exit_activate',
        'on_sbackupConfApp_delete_event',
        'on_sbackupConfApp_destroy'
        ]
    return handlers


def get_restoregui_widgets():
    widget_list = [
        'restorewindow',
        'defaultfolderlabel',
        'defaultradiob',
        'customradiob',
        'custominfos',
        'customentry',
        'customchooser',
        'customapply',
        'calendar',
        'snplist',
        'scrolledwindow1',
        'snplisttreeview',
        'snpdetails',
        'scrolledwindow2',
        'restoreExpander',
        'snpmanExpander',
        'filelisttreeview',
        'buttonspool',
        'restore',
        'restoreas',
        'revert',
        'revertas',
        'txt_current_base',
        'deleteBox',
        'deleteButton',
        'snphistoryFrame',
        'historytv',
        'statusbar',
        'progressbar',
        'box_management',
        'box_restorewin_content',
        'vbox_customdest'
        ]
    return widget_list


def get_restoregui_handlers():
    handlers = [
        'gtk_main_quit',
        'on_defaultradiob_toggled',
        'on_customchooser_clicked',
        'on_customapply_clicked',
        'on_calendar_month_changed',
        'on_calendar_day_selected',
        'on_snplisttreeview_cursor_changed',
        'on_filelisttreeview_row_expanded',
        'on_filelisttreeview_cursor_changed',
        'on_filelisttreeview_unselect_all',
        'on_restore_clicked',
        'on_restoreas_clicked',
        'on_revert_clicked',
        'on_revertas_clicked',
        'on_restoreExpander_activate',
        'on_snpmanExpander_activate',
        'on_deleteButton_clicked',
        'on_exportmanExpander_activate'
        ]
    return handlers




#   Simple Backup - SSH fuse plugin 
#
#   Copyright (c)2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2008: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

from gettext import gettext as _
import re


from sbackup.fuse_plugins import pluginFAM
from sbackup.util.exceptions import FuseFAMException, SBException
from sbackup.util import local_file_utils
from sbackup.util import system

try:
    import pexpect
except ImportError:
    raise FuseFAMException("Unable to import required module `pexpect`")


ssh_re = "^ssh://"
ssh_url_re = ssh_re + "([^:]+?)(:([^@]+?))?@([^/^:^@]+?)(:([0-9]+?))?/(.*)"

class sshFuseFAM (pluginFAM)  :
    """
    The fuseFAM plugin for ssh
    @requires: sshfs, python-pexpect
    @author: Oumar Aziz Ouattara
    """
    def __init__(self):
        pluginFAM.__init__(self)

    def match_scheme(self, remoteSource):
        """
        SSH schema is like : ssh://user:pass@example.com/home/user/backup/ 
        (user,pass, the first '/' ) are mandatory
        """
        _res = False
        _search_res = re.compile(ssh_re).search(remoteSource)
        if _search_res is not None:
            _res = True
        return _res

    def match_scheme_full(self, remoteSource):
        """
        SSH schema is like : ssh://user:pass@example.com/home/user/backup/ 
        (user,pass, the first '/' ) are mandatory
        """
        _res = False
        _search_res = re.compile(ssh_url_re).search(remoteSource)
        if _search_res is not None:
            _res = True
        return _res

    def mount(self, source, mountbase):
        """
        Mount the source intor the mountbase dir . This method should create a mount point to mount the source. 
        The name of the mount point should be very expressive so that we avoid collision with other mount points
        @param source: The remote path
        @param mountbase: The mount points base dir
        @return: The mount point complete path
        """
        exp = re.compile(ssh_url_re)
        match = exp.search(source)
        if not match :
            raise FuseFAMException(_("Error matching the schema 'ssh://user:pass@example.com/home/' with '%s' (The '/' after server is mandatory)") % source)
        else :
            remoteSource = "ssh://" + match.group(1)
            if match.group(3):
                remoteSource += ":" + match.group(3)
            remoteSource += "@" + match.group(4)
            if match.group(6):
                remoteSource += ":" + match.group(6)
            remoteSource += "/"

            user = match.group(1)
            mountpoint = local_file_utils.joinpath(mountbase, self._defineMountDirName(source))
            if match.group(7) :
                pathinside = match.group(7)
            else :
                pathinside = ""

        #If the path is already mounted No need to retry
        if self.checkifmounted(source, mountbase) :
            return (remoteSource, mountpoint, pathinside)

        cmd = "sshfs " + user + "@" + match.group(4) + ":/"
        cmd = cmd + " " + mountpoint

        port = match.group(6)
        if port:
            cmd += " -p " + port
        if not local_file_utils.path_exists(mountpoint) :
            local_file_utils.makedir(mountpoint)

        if system.is_superuser():
            cmd += " -o allow_root"

        self.logger.debug("Spawning: " + cmd)
        password = match.group(3)
        sshfsp = pexpect.spawn(cmd)
        i = sshfsp.expect(['(yes/no)', 'password:', 'Password:', pexpect.EOF])

        if i == 0:
            self.logger.info("Accepting to store the key.")
            sshfsp.sendline('yes')
            i = sshfsp.expect(['(yes/no)', 'password:', 'Password:', pexpect.EOF])

        if i == 1 or i == 2:
            self.logger.debug("Expecting password.")
            if not password :
                sshfsp.sendline("fake")
                local_file_utils.delete(mountpoint)
                raise SBException("sshfs is requesting a password and none has been passed.")
            sshfsp.sendline(password)
            i = sshfsp.expect(['(yes/no)', 'password:', 'Password:', pexpect.EOF])

        result = sshfsp.before # print out the result

        if sshfsp.isalive() or sshfsp.exitstatus:
            local_file_utils.delete(mountpoint)
            raise SBException (_("The sshfs command '%(command)s' didn't perform normally. Output => %(erroroutput)s ") % {"command" : cmd, "erroroutput" : result})

        return (remoteSource, mountpoint, pathinside)

    def getdoc(self):
        doc = _("SSH schema is like : ssh://user:pass@example.com:33/home/user/backup/")
        return doc

    def checkifmounted(self, source, mountbase):
        """
        @return: True if mounted, False if not
        """
        mountpoint = local_file_utils.joinpath(mountbase, self._defineMountDirName(source))
        return local_file_utils.is_mount(mountpoint)

    def _defineMountDirName(self, remote):
        """
        """
        exp = re.compile(ssh_url_re)
        match = exp.search(remote)
        if not match :
            raise FuseFAMException(_("Error matching the schema 'ssh://user:pass@example.com:21/home/' with '%s' (The '/' after server is mandatory)") % remote)
        else :
            user = match.group(1)
            dirname = "ssh_" + user + "@" + match.group(4)
            if match.group(6):
                dirname += "_" + match.group(6)
            return dirname



#   Simple Backup - file access management using GIO/GVFS
#
#   Copyright (c)2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`sbackup.util.gio_utils` -- file access management using GIO/GVFS
========================================================================

.. module:: gio_utils
   :synopsis: public entry point to file access management using GIO/GVFS
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>

"""

from gettext import gettext as _
import types
import time
import uuid
import pickle

import glib
import gio


from sbackup.util import local_file_utils
from sbackup.util import exceptions
from sbackup.util import interfaces
from sbackup.util import constants
from sbackup.util import pathparse
from sbackup.util import structs
from sbackup.util import system
from sbackup.util import log

errorcodes = {
    gio.ERROR_ALREADY_MOUNTED : exceptions.ErrorDescription(gio.ERROR_ALREADY_MOUNTED,
                                                            "ERROR_ALREADY_MOUNTED",
                                                            "File is already mounted."),
    gio.ERROR_BUSY : exceptions.ErrorDescription(gio.ERROR_BUSY, "ERROR_BUSY", "File is busy."),
    gio.ERROR_CANCELLED : exceptions.ErrorDescription(gio.ERROR_CANCELLED, "ERROR_CANCELLED",
                                                      "Operation was cancelled. See gio.Cancellable."),
    gio.ERROR_CANT_CREATE_BACKUP : exceptions.ErrorDescription(gio.ERROR_CANT_CREATE_BACKUP,
                                                               "ERROR_CANT_CREATE_BACKUP",
                                                               "Backup couldn't be created."),
    gio.ERROR_CLOSED : exceptions.ErrorDescription(gio.ERROR_CLOSED, "ERROR_CLOSED", "File was closed."),
    gio.ERROR_EXISTS : exceptions.ErrorDescription(gio.ERROR_EXISTS, "ERROR_EXISTS",
                                                   "File already exists error."),
    gio.ERROR_FAILED : exceptions.ErrorDescription(gio.ERROR_FAILED, "ERROR_FAILED",
                                                   "Generic error condition for when any operation fails."),
    gio.ERROR_FAILED_HANDLED : exceptions.ErrorDescription(gio.ERROR_FAILED_HANDLED, "ERROR_FAILED_HANDLED",
            "Operation failed and a helper program has already interacted with the user. Do not display any error dialog."),
    gio.ERROR_FILENAME_TOO_LONG : exceptions.ErrorDescription(gio.ERROR_FILENAME_TOO_LONG,
                                                              "ERROR_FILENAME_TOO_LONG",
                                                              "Filename is too many characters."),
    gio.ERROR_HOST_NOT_FOUND : exceptions.ErrorDescription(gio.ERROR_HOST_NOT_FOUND, "ERROR_HOST_NOT_FOUND",
                                                           "Host couldn't be found (remote operations)."),
    gio.ERROR_INVALID_ARGUMENT : exceptions.ErrorDescription(gio.ERROR_INVALID_ARGUMENT, "ERROR_INVALID_ARGUMENT", "Invalid argument."),
    gio.ERROR_INVALID_FILENAME : exceptions.ErrorDescription(gio.ERROR_INVALID_FILENAME, "ERROR_INVALID_FILENAME", "Filename is invalid or contains invalid characters."),
    gio.ERROR_IS_DIRECTORY : exceptions.ErrorDescription(gio.ERROR_IS_DIRECTORY, "ERROR_IS_DIRECTORY", "File is a directory error."),
    gio.ERROR_NOT_DIRECTORY : exceptions.ErrorDescription(gio.ERROR_NOT_DIRECTORY, "ERROR_NOT_DIRECTORY", "File is not a directory."),
    gio.ERROR_NOT_EMPTY : exceptions.ErrorDescription(gio.ERROR_NOT_EMPTY, "ERROR_NOT_EMPTY", "File is a directory that isn't empty."),
    gio.ERROR_NOT_FOUND : exceptions.ErrorDescription(gio.ERROR_NOT_FOUND, "ERROR_NOT_FOUND", "File not found error."),
    gio.ERROR_NOT_MOUNTABLE_FILE : exceptions.ErrorDescription(gio.ERROR_NOT_MOUNTABLE_FILE, "ERROR_NOT_MOUNTABLE_FILE", "File cannot be mounted."),
    gio.ERROR_NOT_MOUNTED : exceptions.ErrorDescription(gio.ERROR_NOT_MOUNTED, "ERROR_NOT_MOUNTED", "File isn't mounted."),
    gio.ERROR_NOT_REGULAR_FILE : exceptions.ErrorDescription(gio.ERROR_NOT_REGULAR_FILE, "ERROR_NOT_REGULAR_FILE", "File is not a regular file."),
    gio.ERROR_NOT_SUPPORTED : exceptions.ErrorDescription(gio.ERROR_NOT_SUPPORTED, "ERROR_NOT_SUPPORTED", "Operation not supported for the current backend."),
    gio.ERROR_NOT_SYMBOLIC_LINK : exceptions.ErrorDescription(gio.ERROR_NOT_SYMBOLIC_LINK, "ERROR_NOT_SYMBOLIC_LINK", "File is not a symbolic link."),
    gio.ERROR_NO_SPACE : exceptions.ErrorDescription(gio.ERROR_NO_SPACE, "ERROR_NO_SPACE", "No space left on drive."),
    gio.ERROR_PENDING : exceptions.ErrorDescription(gio.ERROR_PENDING, "ERROR_PENDING", "Operations are still pending."),
    gio.ERROR_PERMISSION_DENIED : exceptions.ErrorDescription(gio.ERROR_PERMISSION_DENIED, "ERROR_PERMISSION_DENIED", "Permission denied."),
    gio.ERROR_READ_ONLY : exceptions.ErrorDescription(gio.ERROR_READ_ONLY, "ERROR_READ_ONLY", "File is read only."),
    gio.ERROR_TIMED_OUT : exceptions.ErrorDescription(gio.ERROR_TIMED_OUT, "ERROR_TIMED_OUT", "Operation timed out."),
    gio.ERROR_TOO_MANY_LINKS : exceptions.ErrorDescription(gio.ERROR_TOO_MANY_LINKS, "ERROR_TOO_MANY_LINKS", "File contains too many symbolic links."),
    gio.ERROR_TOO_MANY_OPEN_FILES : exceptions.ErrorDescription(gio.ERROR_TOO_MANY_OPEN_FILES, "ERROR_TOO_MANY_OPEN_FILES", "The current process has too many files open and can't open any more. Duplicate descriptors do count toward this limit. Since 2.20"),
    gio.ERROR_WOULD_BLOCK : exceptions.ErrorDescription(gio.ERROR_WOULD_BLOCK, "ERROR_WOULD_BLOCK", "Operation would block."),
    gio.ERROR_WOULD_MERGE : exceptions.ErrorDescription(gio.ERROR_WOULD_MERGE, "ERROR_WOULD_MERGE", "Operation would merge files."),
    gio.ERROR_WOULD_RECURSE : exceptions.ErrorDescription(gio.ERROR_WOULD_RECURSE, "ERROR_WOULD_RECURSE", "Operation would be recursive."),
    gio.ERROR_WRONG_ETAG : exceptions.ErrorDescription(gio.ERROR_WRONG_ETAG, "ERROR_WRONG_ETAG", "File's Entity Tag was incorrect.")
}

MSG_UNKNOWN_ERROR_CODE = _("Unknown error code:")


MAX_NUMBER_ASK_PASSWORD = 4


REMOTE_SERVICE_SFTP = 101
REMOTE_SERVICE_FTP = 102
#REMOTE_SERVICE_SECURE_WEBDAV = 103
REMOTE_SERVICE_SSH = 104
REMOTE_SERVICE_NFS = 105
REMOTE_SERVICE_SMB = 106


REMOTE_SERVICES_AVAIL = { REMOTE_SERVICE_SFTP : _("SFTP"),
                          REMOTE_SERVICE_SSH : _("SSH"),
                          REMOTE_SERVICE_FTP : _("FTP"),
                          REMOTE_SERVICE_NFS : _("NFS"),
                          REMOTE_SERVICE_SMB : _("SMB")
#                          REMOTE_SERVICE_SECURE_WEBDAV : _("Secure WebDAV (HTTPS)")
                        }
REMOTE_SERVICE_TO_URI_SCHEME = { REMOTE_SERVICE_FTP : pathparse.URI_SCHEME_FTP,
                                 REMOTE_SERVICE_SFTP : pathparse.URI_SCHEME_SFTP,
                                 REMOTE_SERVICE_SSH : pathparse.URI_SCHEME_SSH,
                                 REMOTE_SERVICE_NFS : pathparse.URI_SCHEME_NFS,
                                 REMOTE_SERVICE_SMB : pathparse.URI_SCHEME_SMB
#                                 REMOTE_SERVICE_SECURE_WEBDAV : URI_SCHEME_DAVS
                               }
URI_SCHEME_TO_REMOTE_SERVICE = { REMOTE_SERVICE_TO_URI_SCHEME[REMOTE_SERVICE_FTP] : REMOTE_SERVICE_FTP,
                                 REMOTE_SERVICE_TO_URI_SCHEME[REMOTE_SERVICE_SFTP] : REMOTE_SERVICE_SFTP,
                                 REMOTE_SERVICE_TO_URI_SCHEME[REMOTE_SERVICE_SSH] : REMOTE_SERVICE_SSH,
                                 REMOTE_SERVICE_TO_URI_SCHEME[REMOTE_SERVICE_NFS] : REMOTE_SERVICE_NFS,
                                 REMOTE_SERVICE_TO_URI_SCHEME[REMOTE_SERVICE_SMB] : REMOTE_SERVICE_SMB
#                                 REMOTE_SERVICE_TO_URI_SCHEME[REMOTE_SERVICE_SECURE_WEBDAV] : REMOTE_SERVICE_SECURE_WEBDAV
                               }


class PathInfo(object):
    def __init__(self, path, eff_path, mounted):
        self.path = path
        self.eff_path = eff_path
        self.mounted = mounted

    def __str__(self):
        _res = [ "Path: `%s`" % self.path,
                 "Eff. path: `%s`" % self.eff_path,
                 "Mounted: %s" % self.mounted
               ]
        _res_str = "\n".join(_res)
        return _res_str


#TODO: implement interface?
class GioMountHandler(object):
    """Handles mounting process for a single URI.
    """

    def __init__(self):
        """
        Constructor
        @param configManager: 
        """
        self.__logger = log.LogFactory.getLogger()

        self.__uri = None
        self.__ask_password_cnt = 0

        self.__path_mount_info = None
        self.__mainloop = None
        self.__mount_finish_callback = None
        self.__umount_finish_callback = None

        self._error = None  # latest error during mount/umount

    def use_own_mainloop(self, use = True):
        if use is True:
            if self.__mainloop is None:
                self.__mainloop = glib.MainLoop()
        else:
            self.__mainloop = None

    def set_callbacks(self, mount, umount):
#        self.__logger.debug("set_additional_callbacks - mount: %s - umount: %s" % (str(mount), str(umount)))
        self.__mount_finish_callback = mount
        self.__umount_finish_callback = umount

    def set_uri(self, uri):
        if not isinstance(uri, pathparse.UriParser):
            raise TypeError("Expected URI of type UriParser. Got %s instead." % type(uri))
        self.__uri = uri

    def _get_mount(self, gfileobj):
        _mount = None
        try:
            _mount = gfileobj.find_enclosing_mount(None)
        except gio.Error, error:
            _mount = None
            if error.code == gio.ERROR_NOT_MOUNTED:
                self.__logger.info(_("Unable to get mount: path is not mounted"))
            elif error.code == gio.ERROR_NOT_FOUND:
                self.__logger.info(_("Unable to get mount: path not found when mounting (is probably local)"))
            else:
                self.__logger.info(_("Unable to get mount: %s") % error)
        except glib.GError, error:
            _mount = None
            self.__logger.info(_("Unable to get mount: %s") % error)

        return _mount

    def _is_local(self, gfileobj):
        """
        :todo: Consolidate with `pathparse.is_local`
        """
        _uri_scheme = gfileobj.get_uri_scheme()
        if _uri_scheme == "file":
            _res = True
        else:
            _res = False

        if _res is True:
            self.__logger.debug("Given path is local")
        else:
            self.__logger.debug("Given path is not local")

        return _res

    def _ask_password_cb(self, mount_op, message, default_user, default_domain, flags):
        self.__ask_password_cnt += 1
        self.__logger.debug("ask password (no: %s)" % self.__ask_password_cnt)
        if self.__ask_password_cnt < MAX_NUMBER_ASK_PASSWORD:
            if self.__uri is None:
                raise ValueError("No URI set")
            self.__logger.debug("Setting username: %s password: %s" % (self.__uri.username,
                                                                    "*"*len(self.__uri.password)))
            mount_op.set_username(self.__uri.username)
#            mount_op.set_domain() # required for samba
            # don't set password if none is given (LP #701403)
            if self.__uri.password is not None:
                mount_op.set_password(self.__uri.password)
            mount_op.reply(gio.MOUNT_OPERATION_HANDLED)
        else:
            self._error = exceptions.RemoteMountFailedError("Max. number of password inputs reached. Aborted.")
            mount_op.reply(gio.MOUNT_OPERATION_ABORTED)

    def _set_mount_flag(self, obj, required):
        if self.__uri is None:
            raise ValueError("No URI set")
        path = self.__uri.query_mount_uri()
        self.__path_mount_info = PathInfo(path, obj.get_path(), required)
        self.__logger.debug("Path_mount_info: %s" % str(self.__path_mount_info))

    def _unset_mount_flag(self):
        if self.__uri is None:
            raise ValueError("No URI set")
        if self.__path_mount_info is None:
            self.__logger.warning(_("Unable to un-set mount flag: path is not stored"))
        else:
            self.__path_mount_info = None

    def mount(self):
        if self.__uri is None:
            raise ValueError("No URI set")
        path = self.__uri.query_mount_uri()
        self.__logger.debug("Mount path: %s" % path)
        _gfileobj = gio.File(path)

        _local = self._is_local(_gfileobj)
        if _local:
            self.__logger.debug("Mounting local path is not required - calling additional callback")
#Note: call additional callbacks anyway              
            if self.__mount_finish_callback is not None:
                self.__mount_finish_callback(error = None)
        else:
            self._do_mount(_gfileobj)

    def umount(self, overwrite_require = False):
        if self.__uri is None:
            raise ValueError("No URI set")
        path = self.__uri.query_mount_uri()
        self.__logger.debug("Umount path: %s" % path)
        _gfileobj = gio.File(path)

        _local = self._is_local(_gfileobj)
        if _local:
            self.__logger.debug("Unmounting local path is not required - calling additional callback")
            if self.__umount_finish_callback is not None:
                self.__umount_finish_callback(error = None)
        else:
            try:
                _mounted = self.__path_mount_info.mounted
            except KeyError:
                _mounted = True

            if overwrite_require is True:
                _mounted = True

            if _mounted is True:
                self.__logger.debug("Umount is required")
                self._do_umount(_gfileobj)
            else:
                self.__logger.debug("Umount is not required - calling additional callback")
                if self.__umount_finish_callback is not None:
                    self.__umount_finish_callback(error = None)

    def _do_mount(self, gfileobj):
        op = gio.MountOperation()
        self.__ask_password_cnt = 0
        op.connect('ask-password', self._ask_password_cb)
        try:
            gfileobj.mount_enclosing_volume(op, self._mount_done_cb)
            if self.__mainloop is not None:
                self.__mainloop.run()
        except gio.Error, error:
            self.__logger.error(get_gio_errmsg(error, "Error in `_do_mount`"))
            raise exceptions.RemoteMountFailedError(str(error))
        except glib.GError, error:
            self.__logger.error(str(error))
            raise exceptions.RemoteMountFailedError(str(error))

    def _do_umount(self, gfileobj):
        _mount = self._get_mount(gfileobj)
        error = None
        if _mount is None:
            self.__logger.debug("No mount found.")

            self._unset_mount_flag()
            self._error = error
            if self.__umount_finish_callback is not None:
                self.__logger.debug("Calling additional callback")
                self.__umount_finish_callback(error)

        else:
            try:
                _mount.unmount(self._umount_done_cb, gio.MOUNT_UNMOUNT_FORCE, None, gfileobj)
    #            gobject.timeout_add(15000, self._umount_done_cb, _mount, None, gfileobj)
    # remaining issues:
    # * how to cancel a timeout if mount was successful?
    # * how to return False (quit timer) when calling callback function?
    # * does it work without an already runnning mainloop
                if self.__mainloop is not None:
                    self.__logger.debug("run loop")
                    self.__mainloop.run()
            except gio.Error, error:
                self.__logger.error(get_gio_errmsg(error, "Error in `umount`"))
                raise exceptions.RemoteUmountFailedError(str(error))
            except glib.GError, error:
                self.__logger.error(str(error))
                raise exceptions.RemoteUmountFailedError(str(error))

    def _mount_done_cb(self, obj, res):
        error = None
        try:
            obj.mount_enclosing_volume_finish(res)
            self._set_mount_flag(obj = obj, required = True)
        except gio.Error, error:
            self._set_mount_flag(obj = obj, required = False)
            if error.code == gio.ERROR_ALREADY_MOUNTED:
                self.__logger.info(_("Path is already mounted."))
                error = None
            elif error.code == gio.ERROR_FAILED_HANDLED:
                self.__logger.error(get_gio_errmsg(error, "Mount failed"))
                if self._error is not None:
                    self.__logger.error(str(self._error))
                    error = self._error
            else:
                self.__logger.error(get_gio_errmsg(error, "Error in `_do_mount`"))
        except glib.GError, error:
            self._set_mount_flag(obj = obj, required = False)
            self.__logger.error(str(error))
        finally:
            if self.__mainloop is not None:
                self.__logger.debug("quit loop")
                self.__mainloop.quit()

        self._error = error
        if self.__mount_finish_callback is not None:
            self.__logger.debug("Calling additional callback")
            self.__mount_finish_callback(error)

    def _umount_done_cb(self, obj, res, gfile):
        error = None
        try:
            obj.unmount_finish(res)
        except gio.Error, error:
            if error.code == gio.ERROR_NOT_MOUNTED:
                self.__logger.info(_("Path is not mounted."))
                error = None
            else:
                self.__logger.error(get_gio_errmsg(error, "Error in `_umount_done_cb`"))
        except glib.GError, error:
            self.__logger.error(str(error))
        finally:
            if self.__mainloop is not None:
                self.__logger.debug("quit loop")
                self.__mainloop.quit()

        if error is None:
            self._unset_mount_flag()

        self._error = error
        if self.__umount_finish_callback is not None:
            self.__logger.debug("Calling additional callback")
            self.__umount_finish_callback(error)

    def test_path(self):
        if self.__uri is None:
            raise ValueError("No URI set")

        _mpath = self.__uri.query_mount_uri()

        try:
            _gmpath = gio.File(_mpath)
            _effpath = _gmpath.get_path()
        except (gio.Error, glib.GError), error:
            raise exceptions.RemoteMountTestFailedError(str(error))

        dname = "%s-%s-%s.tmp" % ("sbackup-dir", time.time(), uuid.uuid4())
        tfilen = "%s-%s-%s.tmp" % ("sbackup", time.time(), uuid.uuid4())

        _test_path(_mpath, dname, tfilen)

    def query_fs_info(self):
        if self.__uri is None:
            raise ValueError("No URI set")
        _mpath = self.__uri.query_mount_uri()

        _size = constants.SIZE_FILESYSTEM_UNKNOWN
        _free = constants.FREE_SPACE_UNKNOWN

        _gfo = gio.File(_mpath)
        try:
            _gfoinfo = _gfo.query_filesystem_info("filesystem::*")
        except gio.Error, error:
            self.__logger.warning(get_gio_errmsg(error, "Error in `query_fs_info`"))
        else:
            _size = _gfoinfo.get_attribute_uint64(gio.FILE_ATTRIBUTE_FILESYSTEM_SIZE)
            _free = _gfoinfo.get_attribute_uint64(gio.FILE_ATTRIBUTE_FILESYSTEM_FREE)
        self.__logger.debug("FS info - size: %s free: %s" % (_size, _free))
        return (_size, _free)


class GioOperations(interfaces.IOperations):
    """Collects simple operations    
    """
    __metaclass__ = structs.Singleton

    pathsep = system.PATHSEP

    def __init__(self):
        interfaces.IOperations.__init__(self)

    @classmethod
    def path_exists(cls, path):
        # Be careful: `gfile.query_exists()' returns True even if path is not
        # accessible! Nevertheless we use it here since `gfile.query_info`
        # does not return required attributes from destination directories
        # over ftp to check whether it is actually existing and readable.
        # This caused a regression (LP #1190224) which let the backup fail.
        _gfileobj = gio.File(path)
        _res = _gfileobj.query_exists()
        return _res

    @classmethod
    def openfile_for_write(cls, path):
        _gfileobj = gio.File(path)
#FIXME: etag should be set to None though it doesn't work then!
        _ostr = _gfileobj.replace(etag = '', make_backup = False)
        return _ostr

    @classmethod
    def openfile_for_read(cls, path):
        _gfile = gio.File(path)
        _istr = _gfile.read()
        return _istr

    @classmethod
    def openfile_for_append(cls, path):
        _gfileobj = gio.File(path)
        _ostr = _gfileobj.append_to()
        return _ostr

    @classmethod
    def copyfile(cls, src, dest):
        """Copies given file and metadata (similar to `shutil.copy2`).
        Overwrites `dest` if it already exists
        """
        _src = gio.File(src)
        _dest = gio.File(dest)

        # the source must be a file and exist
        if not cls.path_exists(src):
            raise IOError("Given copy source `%s` does not exist" % _src.get_parse_name())
        if cls.__isfile(_src):
            _src, _dest = cls._prepare_copy(_src, _dest)
            _src.copy(_dest, flags = gio.FILE_COPY_OVERWRITE)
            try:
                _src.copy_attributes(_dest, flags = gio.FILE_COPY_ALL_METADATA)
            except gio.Error:
                raise exceptions.CopyFileAttributesError(\
                            "Unable to copy file attributes (permissions etc.) of file `%s`."\
                            % _src.get_parse_name())
        else:
            _logger = log.LogFactory.getLogger()
            _logger.warning("Given copy source `%s` is not a file. Skipped." % _src.get_parse_name())

    @classmethod
    def _prepare_copy(cls, src_gfile, dst_gfile):
        """Helper function that prepares the given paths for copying
        using 'nssb_copy'.
        
        Source must be a file or symbolic link to a file!
        
        @todo: Implement test case for symbolic links!
        """
        _src_uri = src_gfile.get_uri()
        _dst_uri = dst_gfile.get_uri()

        _src_file = cls.__basename(src_gfile)

        if cls.__isdir(dst_gfile):
            _dstu = cls.joinpath(_dst_uri, _src_file)
            _dst = gio.File(_dstu)
        elif _dst_uri.endswith(cls.pathsep):
            _dstu = cls.joinpath(_dst_uri, _src_file)
            _dst = gio.File(_dstu)
        else:
            _dst = dst_gfile

        _dstdir = cls.get_dirname(_dst.get_uri())
        if cls.path_exists(_dstdir) is False:
            raise IOError("Given copy destination '%s' does not exist" % _dstdir)

        retval = (src_gfile, _dst)

        return retval

    @classmethod
    def _copy_metadata(cls, src, dest):
        _src = gio.File(src)
        _dest = gio.File(dest)
        _src.copy_attributes(_dest, flags = gio.FILE_COPY_ALL_METADATA)

    @classmethod
    def delete(cls, uri):
        """Deletes given file or directory (recursive).
        """
        if (cls.is_dir(uri) is True) and (cls.is_link(uri) is False):
            cls._rmtree_recurse(uri)
        else:
            cls._rm_file(uri)

    @classmethod
    def _rm_file(cls, path):
        # setting of permissions only in case of failures does not work:
        # if directory is read-only we cannot remove files inside even if these are
        # read/write
        try:
            _gfile = gio.File(path)
            _gfile.delete()
        except gio.Error, error:
            raise IOError(str(error))

    @classmethod
    def _rmtree_recurse(cls, path):
        _listing = cls.listdir_fullpath(path)
        for _ent in _listing:
            if (cls.is_dir(_ent) is True) and (cls.is_link(_ent) is False):
                cls._rmtree_recurse(_ent)
            else:
                cls._rm_file(_ent)
        cls._rm_file(path)

    @classmethod
    def _add_write_permission(cls, path, recursive = True):
        """Sets write permissions for user, group, and others for
        given directory or file (recursive). 
        """
        _gfileobj = gio.File(path)
        try:
            _ginfo = _gfileobj.query_info(attributes = gio.FILE_ATTRIBUTE_UNIX_MODE)
            _fmode = _ginfo.get_attribute_uint32(gio.FILE_ATTRIBUTE_UNIX_MODE)
            _new_mode = _fmode | system.UNIX_PERM_ALL_WRITE
            _ginfo.set_attribute_uint32(gio.FILE_ATTRIBUTE_UNIX_MODE, _new_mode)
            _gfileobj.set_attributes_from_info(_ginfo) # setting attributes directly seems broken
        except gio.Error, error:
            _msg = get_gio_errmsg(error, "Unable to set permissions")
            _logger = log.LogFactory.getLogger()
            _logger.warning(_msg)
            return

        if cls.is_dir(path) and recursive is True:
            for _entryp in cls.listdir_fullpath(path):
                if (cls.is_dir(_entryp) is True)  and (cls.is_link(_entryp) is False):
                    cls._add_write_permission(_entryp)
                else:
                    cls._add_write_permission(_entryp, recursive = False)

    @classmethod
    def chmod_no_rwx_grp_oth(cls, path):
        """Sets write permissions for user only for
        given directory or file (*not* recursive). 
        """
        _gfileobj = gio.File(path)
        try:
            _ginfo = _gfileobj.query_info(attributes = gio.FILE_ATTRIBUTE_UNIX_MODE)
            _fmode = _ginfo.get_attribute_uint32(gio.FILE_ATTRIBUTE_UNIX_MODE)
            _new_mode = _fmode & system.UNIX_PERM_GRPOTH_NORWX
            _ginfo.set_attribute_uint32(gio.FILE_ATTRIBUTE_UNIX_MODE, _new_mode)
            _gfileobj.set_attributes_from_info(_ginfo) # setting attributes directly seems broken
        except gio.Error, error:
            _msg = get_gio_errmsg(error, "Unable to set permissions")
            _logger = log.LogFactory.getLogger()
            _logger.warning(_msg)
            return

    @classmethod
    def force_delete(cls, path):
        cls._add_write_permission(path, recursive = True)
        cls.delete(path)

    @classmethod
    def force_move(cls, src, dst):
        _gsrc = gio.File(src)
        _gdst = gio.File(dst)
        try:
            _gsrc.move(_gdst, flags = gio.FILE_COPY_OVERWRITE)
        except gio.Error:
            if cls.is_dir(src):
                cls._copytree(src, dst)
                cls.force_delete(src)
            else:
                cls.copyfile(src, dst)
                cls.force_delete(src)

    @classmethod
    def _copytree(cls, src, dst):
        """mod of `shutil.copytree`. This doesn't fail if the
        directory exists, it copies inside.
    
        :param src: source path for copy operation
        :param dst: destination
        :param symlinks: copy symlinks?
        :type src: string
        :type dst: string
    
        """
        names = cls.listdir(src)
        if not cls.path_exists(dst):
            cls.makedirs(dst)
#        errors = []
        for name in names:
            srcname = cls.joinpath(src, name)
            dstname = cls.joinpath(dst, name)
#            try:
            if cls.is_dir(srcname) and not cls.is_link(srcname):
                cls._copytree(srcname, dstname)
            else:
                cls.copyfile(srcname, dstname)
#            except gio.Error, why:
#                errors.append((srcname, dstname, str(why)))
#            # catch the Error from the recursive copytree so that we can
#            # continue with other files
#            except shutil.Error, err:
#                errors.extend(err.args[0])
#        try:
#            shutil.copystat(src, dst)
#        except OSError, why:
#            errors.extend((src, dst, str(why)))
#        if len(errors) > 0:
#            raise shutil.Error, errors
        cls._copy_metadata(src, dst)


    @classmethod
    def is_link(cls, path):
        _res = False
        _gfileobj = gio.File(path)
#        _ftype = _gfileobj.query_file_type() #flags = gio.FILE_QUERY_INFO_NONE, cancellable = None)
        _ftype = cls._query_file_type(_gfileobj)
        if _ftype == gio.FILE_TYPE_SYMBOLIC_LINK:
            _res = True
        return _res

    @classmethod
    def __isdir(cls, gfile):
        """Private helper method that takes `gio.File` rather pathname.
        """
        _res = False
        _ftype = cls._query_file_type(gfile)
        if _ftype == gio.FILE_TYPE_DIRECTORY:
            _res = True
        return _res

    @classmethod
    def __isfile(cls, gfile):
        """Private helper method that takes `gio.File` rather pathname.
        """
        _res = False
        _ftype = cls._query_file_type(gfile)
        if _ftype == gio.FILE_TYPE_REGULAR:
            _res = True
        return _res

    @classmethod
    def _query_file_type(cls, gfile):
        try:
            _info = gfile.query_info(attributes = "standard::type", flags = gio.FILE_QUERY_INFO_NONE,
                                       cancellable = None)
            _ftype = _info.get_file_type()
        except gio.Error, error:
            if error.code == gio.ERROR_NOT_FOUND:
                _ftype = None
            else:
                raise
        return _ftype

    @classmethod
    def test_dir_access(cls, path):
        _gfileobj = gio.File(path)
        try:
            _gfileobj.enumerate_children('standard::name')
        except gio.Error, error:
            raise exceptions.FileAccessException(get_gio_errmsg(error,
                                        "Unable to list directory content"))

    @classmethod
    def is_dir(cls, path):
        _gfileobj = gio.File(path)
        _res = cls.__isdir(_gfileobj)
        return _res

    @classmethod
    def listdir(cls, path) :
        """List a directory. Returns basenames of entries.
        """
        listing = []
        _gfileobj = gio.File(path)
        try:
            _infos = _gfileobj.enumerate_children('standard::name')
        except gio.Error, error:
            if error.code == gio.ERROR_NOT_DIRECTORY:
                _msg = get_gio_errmsg(error, "Unable to list directory content")
                _logger = log.LogFactory.getLogger()
                _logger.warning(_msg)
                _ftype = cls._query_file_type(_gfileobj)
                if _ftype == gio.FILE_TYPE_DIRECTORY:
                    _msg = "Directory"
                elif _ftype == gio.FILE_TYPE_MOUNTABLE:
                    _msg = "Mountable"
                elif _ftype == gio.FILE_TYPE_REGULAR:
                    _msg = "Regular file"
                elif _ftype == gio.FILE_TYPE_SHORTCUT:
                    _msg = "Shortcut"
                elif _ftype == gio.FILE_TYPE_SPECIAL:
                    _msg = "Special file"
                elif _ftype == gio.FILE_TYPE_SYMBOLIC_LINK:
                    _msg = "Symbolic link"
                elif _ftype == gio.FILE_TYPE_UNKNOWN:
                    _msg = "unknown"
                else:
                    _msg = "unknown (no match)"
                _logger.warning("Filetype of `listdir` parameter is: %s" % _msg)
            else:
                raise
        else:
            for _info in _infos:
                listing.append(_info.get_name())
        return listing

    @classmethod
    def listdir_fullpath(cls, path) :
        """List a directory. Returns full paths to entries.
        """
        _lst = cls.listdir(path)
        _res = []
        for _ent in _lst:
            _res.append(cls.joinpath(path, _ent))
        return _res

    @classmethod
    def makedir(cls, path):
        _gfileobj = gio.File(path)
        _gfileobj.make_directory()

    @classmethod
    def makedirs(cls, path):
        _gfileobj = gio.File(path)
        _gfileobj.make_directory_with_parents(gio.Cancellable())

    @classmethod
    def normpath(cls, *args):
        """
        :todo: Implement GIO method!
        """
        return local_file_utils.normpath(*args)

    @classmethod
    def joinpath(cls, *args):
        return pathparse.joinpath(*args)

    @classmethod
    def get_eff_path(cls, path):
        _logger = log.LogFactory.getLogger()
        _logger.debug("get effective path for URI: `%s`" % path)
        _gfile = gio.File(path)
        _eff_path = _gfile.get_path()
        _logger.debug("Effective path: `%s`" % _eff_path)
        return _eff_path

    @classmethod
    def get_dirname(cls, path):
#        _gfileobj = gio.File(path)
#        return _gfileobj.get_parent()
        return local_file_utils.get_dirname(path)

    @classmethod
    def __basename(cls, gfile):
        return gfile.get_basename()

    @classmethod
    def get_basename(cls, path):
        _gfileobj = gio.File(path)
        return _gfileobj.get_basename()

    @classmethod
    def pickleload(cls, path):
        """
        Load a python object from the given pickle file
        @param file: the path of the pickle file
        """
        _str = cls.openfile_for_read(path)
        _pobj = pickle.load(_str)
        _str.close()
        return _pobj

    @classmethod
    def pickledump(cls, datas, path):
        _ostr = cls.openfile_for_write(path)
        pickle.dump(datas , _ostr)
        _ostr.close()

    @classmethod
    def rename(cls, src, dst):
        _gfileobj = gio.File(src)
        _info = _gfileobj.query_info('standard::name')
        _dst = cls.get_basename(dst)
        _gfileobj.set_display_name(_dst)

    @classmethod
    def readfile(self, path):
        """Returns content of file en bloc.
        :return: string
        """
        _gfileobj = gio.File(path)
        _cont_t = _gfileobj.load_contents()   # returns (content, length, etag)
        _cont = _cont_t[0]
        return _cont

    @classmethod
    def writetofile(cls, path, content):
        _ostr = cls.openfile_for_write(path)
        _ostr.write(content)
        _ostr.close()

    @classmethod
    def close_stream(cls, file_desc):
        try:
            file_desc.close()
        except gio.Error, error:
            if error.code == gio.ERROR_CLOSED:
                raise exceptions.FileAlreadyClosedError(_("Error while closing stream: %s") % error)
            else:
                raise exceptions.FileAccessException(get_gio_errmsg(error, _("Error while closing stream")))

def get_gio_errmsg(error, title):
    _msg = "%s: %s" % (title, error)
    try:
        _error_descr = errorcodes[error.code]
        _msg = "%s [%s - %s]" % (_msg, _error_descr.name, _error_descr.message)
    except KeyError:
        _msg = "%s [%s %s]" % (_msg, MSG_UNKNOWN_ERROR_CODE, error.code)
    return _msg


def _test_path(path, testdir_name, testfile_name):
    __logger = log.LogFactory().getLogger()
    _mpath = path

    testdir = GioOperations.joinpath(_mpath, testdir_name)
    testfile = GioOperations.joinpath(testdir, testfile_name)

    __logger.info("Perform tests at specified location")
    try:
        # test specified path
        __logger.debug(_("test specified path for existence using GIO"))
        _gmpath = gio.File(_mpath)
        _exists = _gmpath.query_exists()
        if bool(_exists) is False:
            raise exceptions.RemoteMountTestFailedError("Specified remote path does not exists.")

        # test directory
        __logger.debug("Test testdir: %s" % testdir)
        _gtdir = gio.File(testdir)
        _exists = _gtdir.query_exists()
        if bool(_exists) is True:
            raise exceptions.RemoteMountTestFailedError("Unable to create directory for testing purpose: Directory already exists.")

        __logger.debug(_("Create testdir"))
        _res = _gtdir.make_directory()
        if bool(_res) is False:
            raise exceptions.RemoteMountTestFailedError("Unable to create directory for testing purpose.")

        __logger.debug(_("Test testfile for existence"))
        _gtfile = gio.File(testfile)
        _exists = _gtfile.query_exists()
        if bool(_exists) is True:
            raise exceptions.RemoteMountTestFailedError("Unable to create file for testing purpose: File already exists.")

        _buffer = "Some arbitrary content: %s" % uuid.uuid4()
        __logger.debug(_("Create file"))
        _ostr = _gtfile.create()
        __logger.debug("Write buffer: `%s` to file" % _buffer)
        _ostr.write(_buffer)
        _ostr.close()

        # and re-read
        __logger.debug(_("Re-read test file"))
        _gtfile = gio.File(testfile)
        _exists = _gtfile.query_exists()
        if bool(_exists) is False:
            raise exceptions.RemoteMountTestFailedError("Unable to open file for testing purpose: File does not exists.")
        _cont = _gtfile.load_contents()
        assert len(_cont) == 3
        __logger.debug(_cont[0])
        if _cont[0] != _buffer:
            raise exceptions.RemoteMountTestFailedError("Unable to read content from test file: content differs.")

        # clean-up
        __logger.debug(_("Remove file"))
        _gtfile.delete()
        __logger.debug(_("Remove dir"))
        _gtdir.delete()

    except (gio.Error, glib.GError), error:
        raise exceptions.RemoteMountTestFailedError(str(error))


def get_scheme_from_service(service):
    if not isinstance(service, types.IntType):
        raise TypeError
    if service not in REMOTE_SERVICES_AVAIL:
        raise ValueError("Given remote service not supported")
    if service not in REMOTE_SERVICE_TO_URI_SCHEME:
        raise ValueError("Given remote service not supported")

    _scheme = REMOTE_SERVICE_TO_URI_SCHEME[service]
    return _scheme


def get_service_from_scheme(scheme):
    try:
        _service = URI_SCHEME_TO_REMOTE_SERVICE[scheme]
    except KeyError:
        _service = None
    return _service



#! /usr/bin/env python

#   Simple Backup - post-installation script for upgrading Simple Backup
#
#   Copyright (c)2009-2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`sbackupupgrade` --- post-installation script for upgrading Simple Backup
==============================================================================

.. module:: sbackupupgrade
   :synopsis: helper script for upgrading sbackup from older versions
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>

This script provides necessary functionality when upgrading SBackup
to a new version. Purpose of this script is to be run from the Debian
`postinst` script after package update resp. manually after `make install`.

Since the package (at least core) is configured when this script runs,
we can import modules from sbackup here.
"""

import sys
import traceback
import os
import pwd
import ConfigParser
import re
import shutil

from sbackup.pkginfo import Infos
from sbackup.core import ConfigManager

from sbackup.util import system


# definition of error codes
NO_ERRORS = 0
NO_SUPERUSER = 1
GENERAL_ERROR = 9


class _Config(ConfigParser.ConfigParser):
    """A customized ConfigParser for reading and writing of SBackup
    configuration files.
    
    """
    def __init__(self, configfile):
        """Default constructor. Reads the given file into this
        parser.
        
        :param configfile: Full path to the current configuration file
        
        """
        ConfigParser.ConfigParser.__init__(self)
        self._configfile = configfile
        try:
            fobj = file(self._configfile, "r")
        except IOError:
            print "Unable to open `%s` for reading." % str(self._configfile)
        else:
            self.readfp(fobj, self._configfile)
            fobj.close()

    def optionxform(self, option):
        """
        Default behaviour of ConfigParser is to set the option keys to lowercase. 
        by overiding this method, we make it case sensitive. that's really important for dirconfig pathes 
        """
        return str(option)

    def commit_to_disk(self):
        """Writes the current configuration set to the disk. The
        configuration file given to the constructor is used.
        
        """
        try:
            fobj = file(self._configfile, "wb")
        except IOError:
            print "Unable to open `%s` for writing." % str(self._configfile)
        else:
            ConfigParser.ConfigParser.write(self, fobj)
            fobj.close()


class _Settings(object):
    """Class containing constants for upgrading the log option.
    
    These constants cover
    - `__logindefs`: path to file ``login.defs``
    - `__uidmin_name`: name of entry for minimal uid
    - `__uidmax_name`: name of entry for maximal uid
    
    The constants are accessable using the defined classmethods.         
    """
    __logindefs = "/etc/login.defs"
    __uidmin_name = "UID_MIN"
    __uidmax_name = "UID_MAX"

    def __init__(self):
        pass

    @classmethod
    def get_logindefs_path(cls):
        """Returns the path to the file `login.defs`.            
        """
        return cls.__logindefs

    @classmethod
    def get_uidmin_name(cls):
        """Returns the name of the entry for minimal uid.            
        """
        return cls.__uidmin_name

    @classmethod
    def get_uidmax_name(cls):
        """Returns the name of the entry for maximal uid.            
        """
        return cls.__uidmax_name


class _UpgradeAllConffiles(object):
    """This class encapsulates the upgrading of the log options due
    to changes in log file naming in release 0.2.0-RC3. In prior
    releases the log file was named `sbackup.log`. With release
    0.2.0-RC3 this has changed. From now the log file for the default
    profile is named `sbackup.log` and log files for any other
    profiles are named `sbackup-profilename.log`. This was
    neccessary due to problems with identical names of log files.    
    """

    def __init__(self):
        """Constructor of the log option upgrader.        
        """
        self._min_uid = 1000       # fallback values
        self._max_uid = 60000
        self._users = []
        self._configdirs = []

        reexp_templ = "^%s[ \t]+(\d+)$"
        self._reexp_min_uid = re.compile(reexp_templ
                                         % _Settings.get_uidmin_name(),
                                         re.IGNORECASE)
        self._reexp_max_uid = re.compile(reexp_templ
                                         % _Settings.get_uidmax_name(),
                                         re.IGNORECASE)

    def __repr__(self):
        _repr = ["min uid: %s" % self._min_uid,
                 "max uid: %s" % self._max_uid,
                 "users: %s" % self._users,
                 "config dirs: %s" % self._configdirs
               ]
        return "\n".join(_repr)

    def _read_logindefs(self):
        """Reads the lower and upper limit for user ids from
        the `login.defs` file.        
        """
        defspath = _Settings.get_logindefs_path()
        if os.path.isfile(defspath) and os.access(defspath, os.F_OK and os.R_OK):
            eof = False
            try:
                defsfile = file(defspath, "r")
                while not eof:
                    defsline = defsfile.readline()
                    if defsline == "":
                        eof = True
                    else:
                        defsline = defsline.strip()
                        match = self._reexp_min_uid.search(defsline)
                        if match is not None:
                            self._min_uid = int(match.group(1))

                        match = self._reexp_max_uid.search(defsline)
                        if match is not None:
                            self._max_uid = int(match.group(1))
            except IOError:
                print "Error while reading definitions from '%s'. "\
                      "Using defaults." % (defspath)
            else:
                defsfile.close()
        else:
            print "Unable to read definitions from '%s'. "\
                  "Using defaults." % (defspath)

    def _retrieve_users(self):
        """Retrieves all users from the password database that are
        apparently not system services (using the `uid_min` and
        `uid_max` for this). 
        
        """
        self._users = []
        allpw = pwd.getpwall()
        for cpw in allpw:
            try:
                uid = cpw.pw_uid
            except KeyError:
                continue
            if uid >= self._min_uid and uid <= self._max_uid:
                self._users.append(cpw)

    def _make_configdir_list(self):
        """Creates a list containing all basic configuration directories
        of the previously retrieved users. This includes in any case the
        default configuration directory `/etc`.
        
        :note: It is assumed that user configurations are stored in a\
               directory like `~/.config/sbackup`.
        
        :todo: Implement a better way for retrieval of user's confdirs e.g.\
               by reading the users environ!
               
        """
        self._configdirs = []
        # for the superuser
        self._configdirs.append(ConfigManager.\
                                 ConfigManagerStaticData.get_superuser_confdir())

        # for the other users
        for user in self._users:
            try:
                wdir = user.pw_dir
            except KeyError:
                continue
            wdir = os.path.join(wdir,
                    ConfigManager.ConfigManagerStaticData.get_user_confdir_template())
            self._configdirs.append(wdir)



class CopyConf_nssbackup_to_sbackup_011(_UpgradeAllConffiles):

    def __init__(self):
        _UpgradeAllConffiles.__init__(self)

    def _mk_dir(self, dst, src):
        if not os.path.exists(dst):
            print "`%s` is being created" % dst
            os.mkdir(dst)
            _stats = os.stat(src)
            os.chown(dst, _stats.st_uid, _stats.st_gid)
            os.chmod(dst, _stats.st_mode)

    def _copy_default_profile(self):
        for cdir in self._configdirs:
            cdir_nssb = cdir.replace("sbackup", "nssbackup")

            if os.path.isdir(cdir_nssb) and os.access(cdir_nssb, os.F_OK and os.R_OK):
                self._mk_dir(cdir, cdir_nssb)
                cfile_src = os.path.join(cdir_nssb, "nssbackup.conf")
                cfile_dst = os.path.join(cdir, ConfigManager.ConfigManagerStaticData.get_default_conffile())
                self._copy_configfile(cfile_src, cfile_dst)

    def _copy_other_profiles(self):
        """Modifies the configuration files for the other profiles.
        
        """
        for cdir in self._configdirs:
            pdir = os.path.join(cdir, ConfigManager.ConfigManagerStaticData.get_profiles_dir())

            cdir_nssb = cdir.replace("sbackup", "nssbackup")
            pdir_nssb = os.path.join(cdir_nssb, ConfigManager.ConfigManagerStaticData.get_profiles_dir_nssbackup())

            # get the profile directory for current configuration directory
            if os.path.isdir(pdir_nssb) and\
               os.access(pdir_nssb, os.F_OK and os.R_OK):
                # and get the profiles from the profiles directory
                self._mk_dir(pdir, pdir_nssb)
                profiles_nssb = ConfigManager.get_profiles_nssbackup(pdir_nssb)
                for cprof in profiles_nssb:
                    _src = profiles_nssb[cprof][0]
                    _pname = os.path.basename(_src)
                    _dst = os.path.join(pdir, "sbackup%s" % (_pname.lstrip("nssbackup")))
                    self._copy_configfile(_src, _dst)

    def _copy_configfile(self, src, dst):
        """This method modifies a single configuration file, i.e.
        
        * it reads the existing value from the file
        * retrieves the new value under consideration of the profile name
        * writes the new value to the configuration file.
        
        Files that are not readable/writable are skipped.
        
        """
        if os.path.isfile(src) and\
           os.access(src, os.F_OK and os.R_OK):
            print "checking nssbackup config for copy: %s" % src
            if os.path.exists(dst):
                print "   `%s` already exists. Nothing to do." % dst
            else:
                print "   copy nssbackup configuration"
                print "   from `%s`" % (src)
                print "   to   `%s`" % (dst)

                try:
                    shutil.copy2(src, dst)
                    _stats = os.stat(src)
                    os.chown(dst, _stats.st_uid, _stats.st_gid)
                    os.chmod(dst, _stats.st_mode)
                except (OSError, IOError), error:
                    print "failed (%s)" % error

    def do_upgrade(self):
        """Public method that actually processes the upgrade
        consisting of the following steps:
        
        1. Reading the login defaults for determination of non-system users
        2. Retrieve all users on the system
        3. Make a list of all configuration directories for these users
        4. modify the default profile configuration file
        5. modify the configuration files for any other profiles.
        
        An appropriate error code is returned.
        
        """
        self._read_logindefs()
        self._retrieve_users()
        self._make_configdir_list()

        # keep order of method calls due to directory creation
        self._copy_default_profile()
        self._copy_other_profiles()


class UpgradeConfAllProfiles(_UpgradeAllConffiles):
    """This class encapsulates the upgrading of the log options due
    to changes in log file naming in release 0.2.0-RC3. In prior
    releases the log file was named `sbackup.log`. With release
    0.2.0-RC3 this has changed. From now the log file for the default
    profile is named `sbackup.log` and log files for any other
    profiles are named `sbackup-profilename.log`. This was
    neccessary due to problems with identical names of log files.    
    """

    def __init__(self):
        """Constructor of the log option upgrader.        
        """
        _UpgradeAllConffiles.__init__(self)

    def _modify_default_profile(self):
        """Modifies the configuration file for the default profile.
        
        """
        for cdir in self._configdirs:
            cfile = os.path.join(cdir,
                        ConfigManager.ConfigManagerStaticData.get_default_conffile())
            self._modify_configfile(cfile)

    def _modify_other_profiles(self):
        """Modifies the configuration files for the other profiles.
        
        """
        for cdir in self._configdirs:
            pdir = os.path.join(cdir,
                            ConfigManager.ConfigManagerStaticData.get_profiles_dir())
            # get the profile directory for current configuration directory
            if os.path.isdir(pdir) and\
               os.access(pdir, os.F_OK and os.R_OK and os.W_OK):
                # and get the profiles from the profiles directory
                profiles = ConfigManager.get_profiles(pdir)
                for cprof in profiles:
                    cconf = profiles[cprof][0]
                    self._modify_configfile(cconf)

    def _modify_configfile(self, conffile):
        """This method modifies a single configuration file, i.e.
        
        * it reads the existing value from the file
        * retrieves the new value under consideration of the profile name
        * writes the new value to the configuration file.
        
        Files that are not readable/writable are skipped.
        
        """
        if os.path.isfile(conffile) and\
           os.access(conffile, os.F_OK and os.R_OK and os.W_OK):
            print "checking file: %s" % conffile
            config = _Config(conffile)
            _default_config = ConfigManager.get_default_config_obj()
            if config.has_section("log"):
                if config.has_option("log", "file"):
                    logfile = config.get("log", "file")
                    logdir = os.path.dirname(logfile)

                    new_logfn = ConfigManager.get_logfile_name_template(conffile)
                    new_log = os.path.join(logdir, new_logfn)

                    if logfile == new_log:
                        print "   nothing to do. skipped"
                    else:
                        print "   changing log file option"
                        print "   from `%s`" % (logfile)
                        print "   to   `%s`" % (new_log)
                        config.set("log", "file", str(new_log))
                else:
                    config.set("log", "file", os.path.join(_default_config.get_logdir(),
                                                           ConfigManager.get_logfile_name_template(conffile)))
                    print "   no log file specified. Default value set"

                if not config.has_option("log", "level"):
                    config.set("log", "level", _default_config.get_loglevel())
                    print "   no log level specified. Default value set"
            else:
                config.add_section("log")
                config.set("log", "file", os.path.join(_default_config.get_logdir(),
                                                       ConfigManager.get_logfile_name_template(conffile)))
                config.set("log", "level", _default_config.get_loglevel())
                print "   no section `log` found. Default values set"

            config.commit_to_disk()

            if config.has_section("general"):
                # remove old backuplinks options - it's now always done.
                if config.has_option("general", "backuplinks"):
                    print "   Removing backuplinks option (not needed anymore)"
                    config.remove_option("general", "backuplinks")
                    config.commit_to_disk()

    def do_upgrade(self):
        """Public method that actually processes the upgrade
        consisting of the following steps:
        
        1. Reading the login defaults for determination of non-system users
        2. Retrieve all users on the system
        3. Make a list of all configuration directories for these users
        4. modify the default profile configuration file
        5. modify the configuration files for any other profiles.
        
        An appropriate error code is returned.
        
        """
        try:
            self._read_logindefs()
            self._retrieve_users()
            self._make_configdir_list()

            self._modify_default_profile()
            self._modify_other_profiles()
        except Exception, error:
            print "Error while upgrading profiles: %s" % error
            traceback.print_exc()


class CronSetter(object):
    """Reads schedule info from superuser's
    default profile and write according cron entries.
    
    Purpose is to re-create cron entries after package upgrades
    (i.e. when configurations already exist).
    """

    def __init__(self):
        pass

    def do_upgrade(self):
        """Public method that performs the setting.
        Returns always `NO_ERRORS` as exit code.
        """
        if system.is_superuser():
            try:
                print "Reading schedule info from superuser's default profile"
                _conffilehdl = ConfigManager.ConfigurationFileHandler()
                _defconffile = _conffilehdl.get_default_conffile_fullpath()
                print "Configuration: `%s`" % _defconffile

                if os.path.exists(_defconffile):
                    _conf = ConfigManager.ConfigManager(_defconffile)
                    _conf.write_schedule()
            except Exception, error:
                print "Error while writing CRON settings: %s" % error
                traceback.print_exc()

        else:
            print "Operation requires root privileges"


class UpgradeSBackupConf_v010_011(object):
    """Reads schedule info from superuser's
    default profile and write according cron entries.
    
    Purpose is to re-create cron entries after package upgrades
    (i.e. when configurations already exist).
    """

    def __init__(self):
        self._conffile = None
        self._valid_opts = ConfigManager.ConfigManagerStaticData.get_our_options()

    def _query_default_conffile(self):
        print "Reading schedule info from superuser's default profile"
        _conffilehdl = ConfigManager.ConfigurationFileHandler()
        _defconffile = _conffilehdl.get_default_conffile_fullpath()
        print "Configuration: `%s`" % _defconffile
        self._conffile = _defconffile
        if not os.path.exists(self._conffile):
            self._conffile = None

    def _make_conffile_backup(self):
        _bakf = "%s.before_upgrade_to_v011.bak" % self._conffile
        if not os.path.exists(_bakf):
            shutil.copy2(self._conffile, _bakf)

    def _clean_opts(self):
        print "Validating config file: %s" % self._conffile
        _mod = False
        if (self._valid_opts is None):
            return
        _conf = _Config(self._conffile)
        for section in _conf.sections():
            for key in _conf.options(section):
                if (not self._valid_opts.has_key(section)):
                    _conf.remove_section(section)
                    _mod = True
                    print "Section [%(section)s] in '%(configfile)s' should not exist. Removed."\
                          % {'section': section, 'configfile' : self._conffile}

                if (self._valid_opts[section].has_key(key) or self._valid_opts[section].has_key('*')):
                    continue

                _conf.remove_option(section, key)
                _mod = True
                print "Option '%s' in section '%s' in file '%s' is not known. Removed."\
                      % (key, section, self._conffile)

        if _mod:
            _conf.commit_to_disk()

    def _update_compress_format(self):
        print "Updating config file: %s" % self._conffile
        _conf = _Config(self._conffile)
        if _conf.has_option("general", "format"):
            _compr = _conf.get("general", "format")
            if str(_compr) == "1":  # sbackup < 0.11 compatibility hack
                _compr = "gzip"
                _conf.set("general", "format", _compr)
                _conf.commit_to_disk()

    def _update_schedule(self):
        _conf = ConfigManager.ConfigManager(self._conffile)

        if not _conf.has_section("schedule") \
                or (not _conf.has_option("schedule", "cron") \
                and not _conf.has_option("schedule", "anacron")):
            _sched = _conf.get_schedule_and_probe()
            if _sched is not None:  # entry in cron.* found that is not stored in conf file
                print "Schedule found on FS: %s" % str(_sched)
                _ctype = _sched[0]
                _cexpr = _sched[1]
                assert _ctype in ConfigManager.SCHEDULE_TYPES, "Given schedule type `%s` is invalid" % _ctype
                assert (_cexpr is not None) and (_cexpr != "None"), \
                       "Given cron expression `%s` is invalid" % _cexpr

                _conf.setSchedule(_ctype, _cexpr)
                _conf.saveConf()

    def do_upgrade(self):
        """Public method that performs the setting.
        Returns always `NO_ERRORS` as exit code.
        """
        if system.is_superuser():
            try:
                self._query_default_conffile()
                if self._conffile is not None:
                    self._make_conffile_backup()
                    self._clean_opts()
                    self._update_compress_format()
                    self._update_schedule()
            except Exception, error:
                print "Error while upgrading configuration 0.10 to 0.11: %s" % error
                traceback.print_exc()

        else:
            print "Operation requires root privileges"


class UpgradeApplication(object):
    """The upgrade application class that instantiates several upgrade
    action classes and processes them. Due to this design one can simply
    add and execute further upgrade actions. 
    
    """
    def __init__(self):
        """Default constructor. Creates an `UpgradeLogOption` object.
        
        """
        self.__upgrader_v010 = UpgradeSBackupConf_v010_011()
        self.__upgrader_v011 = UpgradeConfAllProfiles()
        self.__cron_setter = CronSetter()
        self.__nssbackup_configs = CopyConf_nssbackup_to_sbackup_011()

    def main(self):
        """Main method that actually does the upgrade process. It returns
        an appropriate error code. 
        
        """
        print "-" * 60
        print "%s %s upgrade tool" % (Infos.NAME, Infos.VERSION)
        print "-" * 60

        if not system.is_superuser():
            print "Upgrade script must be run with root privileges!"
            retcode = NO_SUPERUSER
        else:
            retcode = NO_ERRORS
            self.__upgrader_v010.do_upgrade()

            self.__nssbackup_configs.do_upgrade()

            self.__upgrader_v011.do_upgrade()
            self.__cron_setter.do_upgrade()

        return retcode


if __name__ == "__main__":
    try:
        _UPGRADER = UpgradeApplication()
        RETC = _UPGRADER.main()
    except:
        print "errors occurred:"
        traceback.print_exc()
        RETC = GENERAL_ERROR

    if RETC == NO_ERRORS:
        print "successful finished."

    sys.exit(RETC)



#   Simple Backup - the actual backup service
#
#   Copyright (c)2008-2010,2013: Jean-Peer Lorenz <peer.loz@gmx.net>
#   Copyright (c)2007-2008: Ouattara Oumar Aziz <wattazoum@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`backupproc` --- the actual backup service
================================================

.. module:: backupproc
   :synopsis: Defines the actual backup service
.. moduleauthor:: Ouattara Oumar Aziz (alias wattazoum) <wattazoum@gmail.com>
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>

"""

from gettext import gettext as _

import traceback
import sys
import smtplib
import socket
import datetime
import optparse
import time

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText


from sbackup.pkginfo import Infos

from sbackup.core.ConfigManager import ConfigManager, get_profiles
from sbackup.core.ConfigManager import ConfigurationFileHandler
from sbackup.core.profile_handler import BackupProfileHandler

from sbackup.util import enable_backup_cancel_signal, enable_termsignal
from sbackup.util import get_resource_file
from sbackup.util import local_file_utils
from sbackup.util import exceptions
from sbackup.util import constants
from sbackup.util import system
from sbackup.util import notifier
from sbackup.util import lock
from sbackup.util import log


def except_hook(etype, evalue, etb):
    _logger = log.LogFactory.getLogger()
    _lines = traceback.format_exception(etype, evalue, etb)
    _lines = "".join(_lines)
    _logger.error("Uncaught exception: %s" % evalue)
    _logger.error(_lines)


sys.excepthook = except_hook
#TODO: move the following into GIO backend: should be executed in case of enabled gio backend only!
system.launch_dbus_if_required()


class SBackupProc(object):
    """This class is intended to be a wrapper of the process of backups of
    multiple profiles.
    
    It manages :
    - the full backup process : creation of instances of the BackupProfileHandler
      with the corresponding config file 
    - the logging of exception not handled by BackupProfileHandler
    - the removal of lockfiles
    - the sending of emails
    
    """

    def __init__(self, notifiers, configfile, dbus_connection = None,
                 use_indicator = False, full_snapshot = False):
        """Default constructor. Basic initializations are done here.

        :param notifiers: instances of notifiers that should be used
        :param configfile: use this configuration file
        :param dbus_connection: DBus connection (connected and registerd)
        
        :type notifiers: list
        
        :note: The configuration managers are retrieved very early
               to ensure that specific logger instances are created.
               
        :todo: Collect options (configfile, use_indicator...) in class/struct!
        
        """
        self.__dbus_conn = dbus_connection
        self.__use_indicator = use_indicator
        self.__full_snp = full_snapshot

        self.__exitcode = constants.EXCODE_GENERAL_ERROR
        self.__errors = []

        # collection of all config managers
        self.__confm = []
        # the name of the currently processed profile
        self.__profilename = None
        self.__retrieve_confm(configfile)

        # here the logger created for the default profile is used
        self.logger = log.LogFactory.getLogger(self.__profilename)
        self.logger.debug("%s %s" % (Infos.NAME, Infos.VERSION))

        # the currently used instance of the BackupProfileHandler
        self.__bprofilehdl = None
        self.__state = notifier.SBackupState()

        self.__notifiers = notifiers
        self.__register_notifiers()
        self.__initialize_notifiers()

    def __register_notifiers(self):
        """Registers the given notifiers as observers.

        :todo: should we give the `state` as parameter to notfier's constructor?

        """
        for _notifier in self.__notifiers:
            self.__state.attach(_notifier)

    def __initialize_notifiers(self):
        """Initializes the given notifiers.

        """
        for _notifier in self.__notifiers:
            _notifier.initialize()

    def __terminate_notifiers(self):
        """Unregisters the given notifiers from state subject.

        """
        for _notifier in self.__notifiers:
            _notifier.publish_exit()
            self.__state.detach(_notifier)

    def __sendEmail(self):
        """Checks if the sent of emails is set in the config file 
        then send an email with the report
        
        :todo: Transfer this functionality to a specialized class!
        
        """
        self.logger.debug("Send email report")
        if self.__bprofilehdl.config.has_option("report", "from") :
            _from = self.__bprofilehdl.config.get("report", "from")
        else :
            hostname = socket.gethostname()
            if "." in hostname :
                mailsuffix = hostname
            else :
                mailsuffix = hostname + ".ext"
            _from = _("SBackup Daemon <%(login)s@%(hostname)s>")\
                    % {'login' : str(system.get_user_from_env()), 'hostname': mailsuffix}

        _to = self.__bprofilehdl.config.get("report", "to")
        _title = _("[SBackup] [%(profile)s] Report of %(date)s")\
                    % { 'profile':self.__profilename,
                        'date': datetime.datetime.now() }
        logf = self.__bprofilehdl.config.get_current_logfile()
        if logf is None:
            _content = _("No log file specified.")
        else:
            if local_file_utils.path_exists(logf):
                _content = local_file_utils.readfile(logf)
            else :
                _content = _("Unable to find log file.")

        server = smtplib.SMTP()
        if self.logger.isEnabledFor(10):
            server.set_debuglevel(True)
        msg = MIMEMultipart()

        msg['Subject'] = _title
        msg['From'] = _from
        msg['To'] = _to
        msg.preamble = _title

        msg_content = MIMEText(_content)
        # Set the filename parameter
        msg_content.add_header('Content-Disposition', 'attachment',
                               filename = "sbackup.log")
        msg.attach(msg_content)

        # getting the connection
        if self.__bprofilehdl.config.has_option("report", "smtpserver") :
            if self.__bprofilehdl.config.has_option("report", "smtpport") :
                server.connect(self.__bprofilehdl.config.get("report", "smtpserver"),
                               self.__bprofilehdl.config.get("report", "smtpport"))
            else :
                server.connect(self.__bprofilehdl.config.get("report", "smtpserver"))
        if self.__bprofilehdl.config.has_option("report", "smtptls") and\
                    self.__bprofilehdl.config.get("report", "smtptls") == "1":
            if self.__bprofilehdl.config.has_option("report", "smtpcert") and\
                    self.__bprofilehdl.config.has_option("report", "smtpkey") :
                server.starttls(self.__bprofilehdl.config.get("report", "smtpkey"),
                                self.__bprofilehdl.config.get("report", "smtpcert"))
            else :
                server.starttls()
        if self.__bprofilehdl.config.has_option("report", "smtpuser") and\
                self.__bprofilehdl.config.has_option("report", "smtppassword") :
            server.login(self.__bprofilehdl.config.get("report", "smtpuser"),
                         self.__bprofilehdl.config.get("report", "smtppassword"))

        # send and close connection
        server.sendmail(_from, _to, msg.as_string())
        server.close()

    def __retrieve_confm(self, force_conffile):
        """Factory method that retrieves the appropriate configuration managers
        for the existing profiles. Super-user rights are taken into account.
        The created configuration managers are stored in member variable
        'self.__confm'.
        
        :todo: Place the path names in class `ConfigStaticData`.
        
        """
        self.__confm = []

        # default profile config file and the config directory is determined
        conffile_hdl = ConfigurationFileHandler()
        if force_conffile is None:
            conffile = conffile_hdl.get_conffile()
        else:
            # conffile given on commandline is treated as default profile's config
            conffile = force_conffile
        confdir = conffile_hdl.get_profilesdir(conffile)

        # create config manager for the default profile and set as current
        if local_file_utils.path_exists(conffile):
            confm = ConfigManager(conffile)
            self.__profilename = confm.getProfileName()
            self.__confm.append(confm)      # store the created ConfigManager in a collection
        else:
            errmsg = _("Critical Error: No configuration file for the default profile was found!\n\nNow continue processing remaining profiles.")
            self.__errors.append(errmsg)

        if force_conffile is None:
            # search for alternate configuration files only if no config file was given
            for _prof in get_profiles(confdir).values():
                _prof_path = _prof[0]
                _prof_enable = _prof[1]

                if _prof_enable:
                    confm = ConfigManager(_prof_path)
                    self.__confm.append(confm)

    def run(self):
        """Actual main method to make backups using SBbackup
        
        - launch BackupProfileHandler with the user configuration file
        - catches all exceptions thrown and logs them
        """
        self.__notify_init_errors()

        for confm in self.__confm:
            try:
                self.__profilename = confm.getProfileName()
                self.logger = log.LogFactory.getLogger(self.__profilename)

                self.__bprofilehdl = BackupProfileHandler(confm, self.__state, self.__dbus_conn,
                                          self.__use_indicator, self.__full_snp)

                self.__write_errors_to_log()

                self.__bprofilehdl.do_hook('pre-backup')
                self.__bprofilehdl.prepare()
                self.__bprofilehdl.process()
                self.__exitcode = self.__bprofilehdl.finish()
                self.__bprofilehdl.do_hook('post-backup')

            except exceptions.BackupCanceledError:
                self.__on_backup_canceled()

            except (SystemExit, KeyboardInterrupt, Exception), error:
                self.__on_backup_error(error)

            self.__on_proc_finish()
            self.__bprofilehdl = None

        self.__terminate_notifiers()
        return self.__exitcode

    def __on_backup_canceled(self):
        try:
            self.logger.warning(_("Backup was canceled by user."))
            self.__state.set_state('backup-canceled')
            if self.__bprofilehdl is not None:
                self.__exitcode = self.__bprofilehdl.cancel()
        except Exception, error:
            self.__notify_error(error, _("An error occurred during the cancellation of backup:"))
            self.__exitcode = constants.EXCODE_GENERAL_ERROR

    def __on_backup_error(self, error):
        """Handles errors that occurs during backup process.
        """
        self.__notify_error(error, _("An error occurred during the backup:"))
        self.__exitcode = constants.EXCODE_BACKUP_ERROR
        if self.__bprofilehdl is not None:
            self.__exitcode = self.__bprofilehdl.finish(error)

        if isinstance(error, (SystemExit, KeyboardInterrupt)):
            raise error

    def __on_proc_finish(self):
        """Method that is finally called after backup process.
        """
        try:
            if self.__bprofilehdl and self.__bprofilehdl.config:
                # send the mail
                if self.__bprofilehdl.config.has_section("report") and self.__bprofilehdl.config.has_option("report", "to") :
                    self.__sendEmail()
        except Exception, error:
            self.__notify_error(error, _("Error when sending email:"))
            self.__exitcode = constants.EXCODE_MAIL_ERROR

    def __notify_error(self, error, title):
        """Sends user notification and logs error that occurred during backup process.
        """
        if self.logger.isEnabledFor(10):
            self.logger.exception("%s\n%s" % (title, error))
        else:
            self.logger.error("%s %s" % (title, error))
        self.__state.set_recent_error(error)
        self.__state.set_state('error')

    def __notify_init_errors(self):
        """Errors that occurred during the initialization process were stored
        in an error list. This error list is showed to the user by this method.
        
        Consolidate this method, `__write_errors_to_log` and `__notify_error`!
        """
        if len(self.__errors) > 0:
            for errmsg in self.__errors:
                self.__state.set_recent_error(errmsg)
                self.__state.set_state('error')

    def __write_errors_to_log(self):
        """Errors that occurred during the initialization process
        were stored in an error list. The full list of errors is
        added to the current log.
        
        """
        if len(self.__errors) > 0:
            self.logger.info(_("The following error(s) occurred before:"))
            for errmsg in self.__errors:
                self.logger.error(errmsg.replace("\n", " "))



class SBackupApp(object):
    """The application that processes the backup.
    
    :todo: Implement a base class providing common commandline parsing etc.!
    
    """

    def __init__(self, argv):
        self.__argv = argv
        self.__lock = lock.ApplicationLock(lockfile = constants.LOCKFILE_BACKUP_FULL_PATH,
                                           processname = constants.BACKUP_COMMAND, pid = system.get_pid())
        self.__options_given = None   # do not modify given options
        self.__use_indicator = True
        self.__dbus_avail = False
        self.__configfile = None

        self.__backupproc = None
        self.__notifiers = []
        # we establish a connection to ensure its presence for progress action
        self.__dbus_conn = None
        self.__exitcode = constants.EXCODE_GENERAL_ERROR

    def create_notifiers(self):
        """Creates notifiers used within the backup process. Note that these
        notifiers are not ready for use yet, they are initialized within
        the backup process. Purpose is to split between notifier creation and
        notifier use: the backup process has no information about the
        specific notifiers (type...)
        :note: import D-Bus related modules only if using D-Bus is enabled        
        """
        if self.__options_given.use_dbus and self.__dbus_avail:
            from sbackup.util import dbus_support
            dbus_notifier = dbus_support.DBusNotifier()
            self.__notifiers.append(dbus_notifier)

    def launch_externals(self):
        if self.__options_given.use_dbus == True:
            self._launch_dbusservice()

        if (self.__use_indicator == True) and (self.__dbus_avail == True):
            try:
                self._launch_indicator()
            except exceptions.ResourceFileNotFoundError:
                self.__use_indicator = False

    def _launch_dbusservice(self):
        """Launches the DBus service and establishes a placeholder
        connection in order to keep the service alive as long as this
        application is running. Call `finalize` to close the
        connection properly when terminating the application.
        
        :note: import D-Bus related modules only if using D-Bus is enabled
        """
        from sbackup.util import dbus_support
        self.__dbus_conn = dbus_support.DBusProviderFacade(constants.BACKUP_PROCESS_NAME)
        try:
            self.__dbus_conn.connect()
            self.__dbus_conn.set_backup_pid(pid = system.get_pid())
            self.__dbus_avail = True
        except exceptions.DBusException:
            print "Unable to launch DBus service"
            self.__dbus_conn = None
            self.__dbus_avail = False
            self.__use_indicator = False

    def _launch_indicator(self):
        """              
        environ: the Gnome-session environ is accessable for root and the
                 user who owns the session.
        :note: import D-Bus related modules only if using D-Bus is enabled                 
        """
        print "Attempt to launch indicator application (status icon)."
        _path_to_app = get_resource_file(constants.INDICATORAPP_FILE)

        session_env = system.get_session_environment()
        if session_env is None:
            print "No desktop session found. Indicator application is not started."
        else:
            _cmd = [_path_to_app]
            if self.__options_given.legacy_appindicator is True:
                _cmd.append("--legacy")

            if not system.is_superuser():
                if system.get_user_from_uid() != session_env["USER"]:
                    _cmd = None
                    print "Unable to launch indicator application as current user.\n"\
                          "You must own current desktop session."

            if _cmd is None:
                print "Unable to launch indicator application"
            else:
                pid = system.exec_command_async(args = _cmd, env = session_env)
                print "Indicator application started (PID: %s)" % pid
                time.sleep(constants.INDICATOR_LAUNCH_PAUSE_SECONDS)


    def finalize(self):
        """Cleaning before terminating the application:
        * disconnects if DBus connection was established.
        """
        if (self.__dbus_avail) and (self.__dbus_conn is not None):
            self.__dbus_conn.quit()
        self.__lock.unlock()
        log.shutdown_logging()


    def parse_cmdline(self):
        """Parses the given commandline options and sets specific
        attributes. It must be considered that the DBus service can
        be used without the tray GUI but in contrast the tray GUI
        cannot be used without the DBus service.
        
        The method uses the commandline arguments given to class'
        constructor.
        
        :todo: An option '--dry-run' would be nice!
        
        """
        usage = "Usage: %prog [options] (use -h or --help for more infos)"
        version = "%prog " + Infos.VERSION
        prog = constants.BACKUP_COMMAND

        parser = optparse.OptionParser(usage = usage, version = version, prog = prog)
        parser.add_option("--no-indicator",
                  action = "store_false", dest = "use_indicator", default = True,
                  help = "don't use the graphical indicator application (status icon)")

        parser.add_option("--no-dbus",
                  action = "store_false", dest = "use_dbus", default = True,
                  help = "don't launch the DBus service and "\
                       "don't use it (implies --no-indicator)")

        parser.add_option("--config-file", dest = "configfile",
                          metavar = "FILE", default = None,
                          help = "set the configuration file to use")

        parser.add_option("--legacy-indicator",
              action = "store_true", dest = "legacy_appindicator", default = False,
              help = "use legacy status icon instead of `libappindicator`")

        parser.add_option("--full",
              action = "store_true", dest = "full_snapshot", default = False,
              help = "create full snapshot")

        (options, args) = parser.parse_args(self.__argv[1:])
        if len(args) > 0:
            parser.error("You must not provide any non-option argument")

        if options.configfile:
            if not local_file_utils.path_exists(options.configfile):
                parser.error("Given configuration file does not exist")
            self.__configfile = options.configfile

        self.__options_given = options

        if self.__options_given.use_dbus == True:
            self.__use_indicator = options.use_indicator
        else:
            self.__use_indicator = False

    def __on_already_running(self, error):
        """Handler for the case a backup process is already running.
        Fuse is not initialized yet.
        :note: import D-Bus related modules only if using D-Bus is enabled        
        """
        print _("Backup is not being started.\n%s") % (str(error))
        if self.__options_given.use_dbus and self.__dbus_avail:
            from sbackup.util import dbus_support
            conn = dbus_support.DBusClientFacade("Simple Backup Process (another instance)")
            conn.connect()
            conn.emit_alreadyrunning_signal()
            conn.quit()
        self.__exitcode = constants.EXCODE_INSTANCE_ALREADY_RUNNING

    def run(self):
        """Runs the whole backup process including launching of
        external applications and services...
        
        """
        try:
            enable_termsignal()
            enable_backup_cancel_signal()
            self.parse_cmdline()

            self.launch_externals()
            self.create_notifiers()

            self.__lock.lock()
            system.very_nice()
            system.set_grp("admin")

            self.__backupproc = SBackupProc(self.__notifiers,
                                              self.__configfile,
                                              self.__dbus_conn,
                                              self.__use_indicator,
                                              self.__options_given.full_snapshot)
            self.__exitcode = self.__backupproc.run()

        except exceptions.InstanceRunningError, error:
            self.__on_already_running(error)
        except SystemExit, exc:
            self.__exitcode = exc.code
        except KeyboardInterrupt:
            self.__exitcode = constants.EXCODE_KEYBOARD_INTERRUPT
        finally:
            self.finalize()

        return self.__exitcode


def main(argv):
    """Public function that process the backups.
    :note: DBus connection is not given as parameter here because it is not clear
           whether to use DBus or not at this time.           
    """
    sbackup_app = SBackupApp(argv)
    exitcode = sbackup_app.run()
    return exitcode



#   Simple Backup - lock file facilities
#
#   Copyright (c)2010: Jean-Peer Lorenz <peer.loz@gmx.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
"""
:mod:`lock` --- lock file facilities
================================================

.. module:: lock
   :synopsis: Provides common lock file facilities
.. moduleauthor:: Jean-Peer Lorenz <peer.loz@gmx.net>

"""

from gettext import gettext as _
import types
import os


from sbackup.util import local_file_utils
from sbackup.util import system
from sbackup.util import exceptions
from sbackup.util import log


class ApplicationLock(object):

    def __init__(self, lockfile, processname, pid):
        if not isinstance(lockfile, types.StringTypes):
            raise TypeError("String for parameter `lockfile`expected")
        if not isinstance(processname, types.StringTypes):
            raise TypeError("String for parameter `processname`expected")
        if not isinstance(pid, types.IntType):
            raise TypeError("String for parameter `pid`expected")

        self.__lockfile = lockfile
        self.__processname = processname
        self.__pid = pid
        self.__logger = log.LogFactory.getLogger()

    def __prepare_lock_dir(self):
        _dir = os.path.dirname(self.__lockfile)
        if not local_file_utils.path_exists(_dir):
            try:
                os.mkdir(_dir)
                os.chmod(_dir, 0777)
            except (OSError, IOError), error:
                self.__logger.error("Unable to make lock directory: %s" % error)
                raise exceptions.ApplicationLockError

    def lock(self):
        """Sets a lock file. 
        In 0.3 following changes take effect:
        * use fixed location (users and superusers); ignore settings in configuration files
        * existence of directory `sbackup` with mode 777 is assumed (no sticky bit set) 
        """
        self.__prepare_lock_dir()

        if local_file_utils.path_exists(self.__lockfile):
            if self.__is_lock_valid() is True:
                raise exceptions.InstanceRunningError(\
                    _("Another application instance is already running."))
            else:
                self.__logger.info(_("Invalid lock file found. Is being removed."))
                self.__force_unsetlock()
        try:
            local_file_utils.writetofile(self.__lockfile, str(self.__pid))
            self.__logger.debug("Created lockfile `%s` with info `%s`." % (self.__lockfile, str(self.__pid)))
        except (OSError, IOError), error:
            self.__logger.error("Unable to create lock: %s" % error)
            raise exceptions.ApplicationLockError

    def __is_lock_valid(self):
        valid = False
        if local_file_utils.path_exists(self.__lockfile):
            # the lockfile exists, is it valid?
            try:
                last_pid = local_file_utils.readfile(self.__lockfile)
                last_pid = int(last_pid)
            except (OSError, IOError, ValueError), error:
                self.__logger.error("Error while reading lockfile: %s" % str(error))
                last_pid = None

            if last_pid is not None:
                if system.pid_exists(pid = str(last_pid), processname = self.__processname):
                    valid = True
        return valid

    def __is_lock_owned(self):
        owned = False
        if local_file_utils.path_exists(self.__lockfile):
            # the lockfile exists, is it valid?
            try:
                last_sb_pid = local_file_utils.readfile(self.__lockfile)
                last_sb_pid = int(last_sb_pid)
            except (OSError, IOError, ValueError), error:
                self.__logger.error("Error while reading lockfile: %s" % str(error))
                last_sb_pid = None

            if last_sb_pid is not None:
                if last_sb_pid == self.__pid:
                    owned = True
        return owned

    def unlock(self):
        """Remove lockfile.
        """
        if local_file_utils.path_exists(self.__lockfile):
            if self.__is_lock_owned():
                try:
                    local_file_utils.delete(self.__lockfile)
                    self.__logger.debug("Lock file '%s' removed." % self.__lockfile)
                except  (OSError, IOError), _exc:
                    self.__logger.error(_("Unable to remove lock file: %s") % str(_exc))
            else:
                self.__logger.debug("Unable to remove lock: not owned by this process.")
        else:
            self.__logger.warning(_("Unable to remove lock file: File not found."))

    def __force_unsetlock(self):
        """Remove lockfile.
        """
        if local_file_utils.path_exists(self.__lockfile):
            try:
                local_file_utils.delete(self.__lockfile)
                self.__logger.debug("Lock file '%s' removed." % self.__lockfile)
            except (OSError, IOError), _exc:
                self.__logger.error(_("Unable to remove lock file: %s") % str(_exc))
        else:
            self.__logger.info(_("Unable to remove lock file: File not found."))
